import os
from logging import Logger
from typing import Optional, Tuple, Dict, Set

import aiofiles

from guapow import __app_name__
from guapow.common import log
from guapow.common.model import FileModel, RootFileModel
from guapow.common.model_util import FileModelFiller
from guapow.common.users import is_root_user

HTTP_SERVER_PORT = 'GUAPOW_OPT_PORT'


def get_root_optimizer_config_path():
    return f'/etc/{__app_name__}/opt.conf'


def get_user_optimizer_config_path(user_name: str):
    return f'/home/{user_name}/.config/{__app_name__}/opt.conf'


def get_optimizer_config_path_by_user_priority(current_user_id: int, current_user_name: str, logger: Logger) -> Optional[str]:
    is_root = is_root_user(current_user_id)
    file_path = get_root_optimizer_config_path() if is_root else get_user_optimizer_config_path(current_user_name)

    if os.path.isfile(file_path):
        logger.info(f"Optimizer configuration file '{file_path}' found")
        return file_path
    else:
        logger.warning(f"Optimizer configuration file '{file_path}' not found")
        return get_optimizer_config_path_by_user_priority(0, 'root', logger) if not is_root else None  # tries to retrieve it from root


class RequestSettings(FileModel):

    FILE_MAPPING = {'allowed_users': ('allowed_users', Set[str], None),
                    'encrypted': ('encrypted', bool, True)}

    def __init__(self, encrypted: Optional[bool], allowed_users: Optional[Set[str]]):
        self.encrypted = encrypted
        self.allowed_users = allowed_users

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        return 'request'

    def is_valid(self) -> bool:
        return self.encrypted is not None

    def setup_valid_properties(self):
        if self.encrypted is None:
            self.encrypted = True

    @classmethod
    def empty(cls) -> Optional["RequestSettings"]:
        return cls(None, None)

    @classmethod
    def default(cls) -> Optional["RequestSettings"]:
        instance = cls.empty()
        instance.setup_valid_properties()
        return instance


class OptimizerConfig(RootFileModel):

    DEFAULT_PORT = 5087

    FILE_MAPPING = {'port': ('port', int, None),
                    'compositor': ('compositor', str, None),
                    'scripts.allow_root': ('allow_root_scripts', bool, True),
                    'check.finished.interval': ('check_finished_interval', float, None),
                    'launcher.mapping.timeout': ('launcher_mapping_timeout', float, None),
                    'gpu.cache': ('gpu_cache', bool, True),
                    'gpu.id': ('gpu_ids', Set[int], None),
                    'gpu.vendor': ('gpu_vendor', str, None),
                    'cpu.performance': ('cpu_performance', bool, True),
                    'profile.cache': ('profile_cache', bool, True),
                    'profile.pre_caching': ('pre_cache_profiles', bool, True),
                    'nice.check.interval': ('renicer_interval', float, None)}

    def __init__(self, port: Optional[int] = None, compositor: Optional[str] = None,
                 allow_root_scripts: Optional[bool] = False,
                 check_finished_interval: Optional[float] = None, launcher_mapping_timeout: Optional[float] = 30,
                 gpu_cache: Optional[bool] = False, cpu_performance: Optional[bool] = None,
                 profile_cache: Optional[bool] = None, pre_cache_profiles: Optional[bool] = None,
                 gpu_vendor: Optional[str] = None, renicer_interval: Optional[float] = None,
                 gpu_ids: Optional[Set[int]] = None):
        self.port = port
        self.compositor = compositor
        self.allow_root_scripts = allow_root_scripts
        self.check_finished_interval = check_finished_interval
        self.launcher_mapping_timeout = launcher_mapping_timeout
        self.gpu_cache = gpu_cache
        self.gpu_vendor = gpu_vendor
        self.gpu_ids = gpu_ids
        self.cpu_performance = cpu_performance
        self.request = RequestSettings.default()
        self.profile_cache = profile_cache
        self.pre_cache_profiles = pre_cache_profiles
        self.renicer_interval = renicer_interval

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        pass

    def is_valid(self) -> bool:
        if super(OptimizerConfig, self).is_valid():
            return True

        return all([self.has_valid_port(),
                    bool(self.compositor),
                    self.allow_root_scripts is not None,
                    self.cpu_performance is not None,
                    self.profile_cache is not None,
                    self.has_valid_check_finished_interval(),
                    self.has_valid_launcher_mapping_timeout(),
                    self.has_valid_renicer_interval()])

    def has_valid_port(self) -> bool:
        return self.port is not None and 0 <= self.port <= 65535

    def has_valid_launcher_mapping_timeout(self) -> bool:
        return self.launcher_mapping_timeout is not None and self.launcher_mapping_timeout >= 0

    def has_valid_check_finished_interval(self) -> bool:
        return self.check_finished_interval is not None and self.check_finished_interval >= 0.5

    def has_valid_renicer_interval(self) -> bool:
        return self.renicer_interval is not None and self.renicer_interval > 0

    def setup_valid_properties(self):
        if self.port is None:
            try:
                self.port = int(os.getenv(HTTP_SERVER_PORT, self.DEFAULT_PORT))
            except ValueError:
                self.port = self.DEFAULT_PORT

        if not self.has_valid_port():
            self.port = self.DEFAULT_PORT

        if not self.has_valid_check_finished_interval():
            self.check_finished_interval = 3

        if not self.has_valid_launcher_mapping_timeout():
            self.launcher_mapping_timeout = 30

        if self.gpu_cache is None:
            self.gpu_cache = False

        if self.profile_cache is None:
            self.profile_cache = False

        if self.pre_cache_profiles is None:
            self.pre_cache_profiles = False

        if not self.has_valid_renicer_interval():
            self.renicer_interval = 5

        if self.request is None:
            self.request = RequestSettings.default()
        elif not self.request.is_valid():
            self.request.setup_valid_properties()

    @property
    def encrypted_requests(self) -> bool:
        return bool(self.request and self.request.encrypted)

    @classmethod
    def read_valid(cls, logger: Logger, model_filler: FileModelFiller, file_path: str, only_properties: Optional[Set[str]] = None) -> "OptimizerConfig":
        instance = cls()

        try:
            with open(file_path) as f:
                config_str = f.read().strip()

            model_filler.fill(instance, config_str, only_properties)
        except FileNotFoundError:
            logger.warning(f"Optimizer configuration file '{file_path}' does not exist. Using default settings.")

        instance.setup_valid_properties()

        if instance.is_valid():
            return instance
        else:
            logger.warning(f"Invalid optimizer configuration file '{file_path}'")

    @staticmethod
    def is_log_enabled() -> bool:
        try:
            return bool(int(os.getenv('GUAPOW_OPT_LOG', '1')))
        except ValueError:
            return True

    @staticmethod
    def get_log_level() -> int:
        return log.get_log_level('GUAPOW_OPT_LOG_LEVEL')

    @staticmethod
    def is_service() -> bool:
        try:
            return bool(int(os.getenv('GUAPOW_OPT_SERVICE', '0')))
        except ValueError:
            return False

    @classmethod
    def empty(cls) -> "OptimizerConfig":
        instance = cls(allow_root_scripts=None, check_finished_interval=None, launcher_mapping_timeout=None,
                       gpu_cache=None)
        instance.request = None
        return instance

    @classmethod
    def default(cls) -> "OptimizerConfig":
        instance = OptimizerConfig()
        instance.setup_valid_properties()
        return instance


class OptimizerConfigReader:

    def __init__(self, model_filler: FileModelFiller, logger: Logger):
        self._model_filler = model_filler
        self._log = logger

    async def read_valid(self, file_path: str, only_properties: Optional[Set[str]] = None) -> Optional[OptimizerConfig]:
        instance = OptimizerConfig()

        try:
            async with aiofiles.open(file_path) as f:
                config_str = (await f.read()).strip()

            self._model_filler.fill(instance, config_str, only_properties)
        except FileNotFoundError:
            self._log.warning(f"Optimizer configuration file '{file_path}' does not exist. Using default settings.")

        instance.setup_valid_properties()

        if instance.is_valid():
            return instance
        else:
            self._log.warning(f"Invalid optimizer configuration file '{file_path}'")


async def read_optimizer_config(user_id: int, user_name: str, filler: FileModelFiller, logger: Logger, only_properties: Optional[Set[str]] = None) -> Optional[OptimizerConfig]:
    opt_config_path = get_optimizer_config_path_by_user_priority(user_id, user_name, logger)

    if opt_config_path:
        return await OptimizerConfigReader(filler, logger).read_valid(file_path=opt_config_path,
                                                                      only_properties=only_properties)

    else:
        logger.warning("Using default Optimizer settings")
        return OptimizerConfig.default()



from logging import Logger

from aiohttp import web

from guapow.common.auth import read_machine_id
from guapow.common.config import OptimizerConfig
from guapow.common.users import map_home_users, map_all_users
from guapow.service.optimizer.flow import OptimizationQueue
from guapow.service.optimizer.handler import OptimizationHandler
from guapow.service.optimizer.web import routes, middlewares


async def create_web_app(handler: OptimizationHandler, queue: OptimizationQueue, config: OptimizerConfig, logger: Logger) -> web.Application:
    app = web.Application()
    app.logger = logger
    app['queue'] = queue
    app['handler'] = handler

    if config.request and config.request.allowed_users:
        all_users = map_all_users()
        app['allowed_users'] = {user: all_users.get(user) for user in config.request.allowed_users if user in all_users}
    else:
        app['allowed_users'] = map_home_users()

    logger.info(f"Requests allowed for users: {', '.join(sorted(app['allowed_users']))}")

    if config.encrypted_requests:
        machine_id = await read_machine_id()

        if machine_id:
            app['machine_id'] = machine_id
            app.middlewares.append(middlewares.decrypt_request)
            logger.info("Requests encryption is enabled")
        else:
            logger.warning("Requests encryption is disabled since machine-id could not be read")
    else:
        logger.warning("Requests encryption is disabled")

    app.middlewares.append(middlewares.hide_server_headers)
    app.add_routes([web.post('/', routes.optimize)])
    return app



import asyncio
from asyncio import Lock
from copy import deepcopy
from typing import Optional, List, Set, Dict

from guapow.common import system
from guapow.service.optimizer.post_process.context import PostProcessContext, PostProcessContextMapper
from guapow.service.optimizer.post_process.summary import GeneralPostProcessSummarizer
from guapow.service.optimizer.post_process.task import PostProcessTaskManager
from guapow.service.optimizer.task.model import OptimizedProcess, OptimizationContext


class DeadProcessWatcher:

    def __init__(self, context: OptimizationContext, restore_man: PostProcessTaskManager, check_interval: float, to_watch: Optional[List[OptimizedProcess]] = None,
                 to_relaunch: Optional[Dict[str, str]] = None):
        self._context = context
        self._check_interval = check_interval
        self._restore_man = restore_man
        self._log = context.logger
        self._to_watch = to_watch
        self._lock_to_watch = Lock()
        self._watching = False
        self._lock_watching = Lock()
        self._to_relaunch = to_relaunch

        if self._to_watch:
            for p in self._to_watch:
                self._register_post_commands_to_relaunch(p)

    def _update_to_relaunch(self, commands: Dict[str, str]):
        for comm, cmd in commands.items():
            cached_cmd = self._to_relaunch.get(comm)

            if not cached_cmd or not cached_cmd.startswith('/'):
                self._to_relaunch[comm] = cmd

    def _register_post_commands_to_relaunch(self, process: OptimizedProcess):
        if process.stopped_processes and process.relaunch_stopped_processes:
            self._update_to_relaunch(process.stopped_processes)

        if process.stopped_after_launch and process.relaunch_stopped_after_launch:
            self._update_to_relaunch(process.stopped_after_launch)

    async def watch(self, process: OptimizedProcess):
        async with self._lock_to_watch:
            if self._to_watch is None:
                self._to_watch = []

            self._to_watch.append(process)
            self._log.debug(f"Watching a new process '{process.pid}' ({len(self._to_watch)} now)")
            self._register_post_commands_to_relaunch(process)

    async def map_context(self) -> PostProcessContext:
        post_summary = await GeneralPostProcessSummarizer.instance().summarize(processes=self._to_watch,
                                                                               pids_alive=system.read_current_pids(),
                                                                               processes_to_relaunch=self._to_relaunch,
                                                                               context=self._context)

        if post_summary.dead_pids:
            await self._context.queue.remove_pids(*(data[1] for data in post_summary.dead_pids))

            dead_pids = sorted(post_summary.dead_pids)
            for idx, data in enumerate(dead_pids):
                del self._to_watch[data[0] - idx]

            self._log.debug(f"{len(dead_pids)} process(es) stopped: {', '.join((str(p[1]) for p in dead_pids))}")

        context = PostProcessContextMapper.instance().map(post_summary)

        if self._to_relaunch:  # cleaning up commands that will be relaunched
            if context.stopped_processes:
                for comm_cmd in context.stopped_processes:
                    if comm_cmd[0] in self._to_relaunch:
                        del self._to_relaunch[comm_cmd[0]]

            if context.not_stopped_processes:
                for comm in context.not_stopped_processes:
                    if comm in self._to_relaunch:
                        del self._to_relaunch[comm]

        return context

    async def start_watching(self):
        async with self._lock_watching:
            self._watching = True

        while True:
            async with self._lock_to_watch:
                context = await self.map_context()

            restore_tasks = self._restore_man.create_tasks(context)

            if restore_tasks:
                await asyncio.gather(*restore_tasks)

            async with self._lock_to_watch:
                if not self._to_watch:
                    break

            await asyncio.sleep(self._check_interval)

        async with self._lock_watching:
            self._watching = False

        self._log.debug("No processes to watch. Stopped watching.")

    async def is_watching(self) -> bool:
        async with self._lock_watching:
            return self._watching

    async def get_watched_pids(self) -> Optional[Set[int]]:
        if self._to_watch:
            async with self._lock_to_watch:
                return {p.pid for p in self._to_watch}

    def get_to_relaunch_view(self) -> Optional[Dict[str, str]]:
        if self._to_relaunch is not None:
            return deepcopy(self._to_relaunch)


class DeadProcessWatcherManager:

    def __init__(self, check_interval: float, restore_man: PostProcessTaskManager, context: OptimizationContext):
        self._watcher = DeadProcessWatcher(context=context,
                                           check_interval=check_interval,
                                           restore_man=restore_man,
                                           to_relaunch={})

    async def watch(self, process: OptimizedProcess):
        if process:
            await self._watcher.watch(process)
            if not await self._watcher.is_watching():
                asyncio.get_event_loop().create_task(self._watcher.start_watching())



import os
import re
import shutil
from logging import Logger
from pathlib import Path
from typing import Dict, Optional, Tuple

from guapow import ROOT_DIR
from guapow.common import system
from guapow.common.users import is_root_user


def get_systemd_root_service_dir() -> str:
    return '/usr/lib/systemd/system'


def get_systemd_user_service_dir() -> str:
    return f'{Path.home()}/.config/systemd/user'


def get_service_dir(root: bool) -> str:
    return get_systemd_root_service_dir() if root else get_systemd_user_service_dir()


def map_service_status(output: str) -> Optional[Tuple[str, str]]:
    res = re.compile(r'\s+loaded\s+\((/.+\.service);\s*(\w+);.*\)').findall(output)
    return res[0] if res else None


def gen_english_environment() -> Dict[str, str]:
    custom_env = dict(os.environ)
    custom_env['LANG'] = 'en_US.UTF-8'
    return custom_env


def get_source_service_path(file_name: str, root: bool) -> str:
    return f"{ROOT_DIR}/dist/daemon/systemd/{'root' if root else 'user'}/{file_name}"


def get_service_status(file_name: str, root: bool, env: Dict[str, str]) -> Optional[str]:
    return system.syscall(f"systemctl status{' --user' if not root else ''} {file_name}", custom_env=env)[1]


def enable_service(file_name: str, root: bool, env: Dict[str, str]) -> Tuple[int, Optional[str]]:
    return system.syscall(f"systemctl enable{' --user' if not root else ''} --now {file_name}", custom_env=env)


def disable_service(file_name: str, root: bool, env: Dict[str, str]) -> Tuple[int, Optional[str]]:
    return system.syscall(f"systemctl disable{' --user' if not root else ''} --now {file_name}", custom_env=env)


class ServiceInstaller:

    def __init__(self, service_cmd: str, service_file: str, requires_root: bool, logger: Logger):
        self._service_cmd = service_cmd
        self._service_file = service_file
        self._requires_root = requires_root
        self._log = logger

    def install(self) -> bool:
        root = is_root_user()

        if self._requires_root and not root:
            self._log.error("Requires root privileges")
            return False

        if not shutil.which('systemctl'):
            self._log.error("'systemctl' is not installed on your system")
            return False

        service_cmd = shutil.which(self._service_cmd)

        if not service_cmd:
            self._log.error(f"'{self._service_cmd}' is not installed on your system")
            return False

        service_dir = get_service_dir(root)
        dest_file = f'{service_dir}/{self._service_file}'

        if not os.path.exists(dest_file):
            try:
                Path(service_dir).mkdir(exist_ok=True, parents=True)
            except OSError:
                self._log.error(f"Could not create directory '{service_dir}'")
                return False

            local_file = get_source_service_path(self._service_file, root)

            if not os.path.exists(local_file):
                self._log.error(f"Service definition file '{local_file}' not found")
                return False

            with open(local_file) as f:
                service_definition = f.read()

            replace_pattern = re.compile(r'ExecStart=.+\n')
            service_definition = replace_pattern.sub(f'ExecStart={service_cmd}\n', service_definition)

            try:
                with open(dest_file, 'w+') as f:
                    f.write(service_definition)
                    
                self._log.info(f"File '{local_file}' copied to '{dest_file}'")
            except OSError:
                self._log.error(f"Could not copy service file '{local_file}' to '{service_dir}'")
                return False

        custom_env = gen_english_environment()
        status_output = get_service_status(self._service_file, root, custom_env)

        status = map_service_status(status_output)

        if status and len(status) != 2:
            self._log.error("Unknown status output from 'systemctl' output: {}".format(status_output.replace('\n', ' ')))
            return False

        if status[1].strip() == 'enabled':
            return True

        enabled_code, enabled_output = enable_service(self._service_file, root, custom_env)

        if enabled_code != 0:
            error_log = ': {}'.format(enabled_output.replace('\n', ' ')) if enabled_output else ''
            self._log.error(f"Could not enable {self._service_file}. Output (exitcode={enabled_code}){error_log}")
            return False

        self._log.info(f"{self._service_file} enabled and started")
        return True


class ServiceUninstaller:

    def __init__(self, service_file: str, requires_root: bool, logger: Logger):
        self._service_file = service_file
        self._requires_root = requires_root
        self._log = logger

    def uninstall(self) -> bool:
        root = is_root_user()

        if self._requires_root and not root:
            self._log.error("Requires root privileges")
            return False

        if not shutil.which('systemctl'):
            self._log.error("'systemctl' is not installed on your system")
            return False

        custom_env = gen_english_environment()

        status_output = get_service_status(self._service_file, root, custom_env)

        if not status_output:
            self._log.error(f"Could not retrieve information of '{self._service_file}'")
            return False

        if ' could not be found' in status_output:
            self._log.info(f"{self._service_file} is not installed")
            return True

        status = map_service_status(status_output)

        if not status or len(status) != 2:
            output_log = status_output.replace('\n', ' ')
            self._log.error(f"Could not determine '{self._service_file}' status on the 'systemctl' output: {output_log}")
            return False

        file_path, service_status = status[0].strip(), status[1].strip()

        if service_status.lower() == 'enabled':
            disable_code, disable_output = disable_service(self._service_file, root, custom_env)

            if disable_code != 0:
                self._log.error(f"Could not disable '{self._service_file}'")
                return False

            self._log.info(f"{self._service_file} stopped and disabled")

        if os.path.exists(file_path):
            try:
                os.remove(file_path)
            except OSError:
                self._log.error(f"Could not remove file '{file_path}'")
                return False

            self._log.info(f"'{file_path}' removed")

        return True



import asyncio
from typing import Dict, Tuple

from guapow.common.system import async_syscall


async def _map_processes_by_attribute(attr: str) -> Dict[int, str]:
    code, output = await async_syscall(f'ps -Ao "%p#%{attr}" -ww --no-headers')

    if code == 0:
        procs = {}
        for line in output.split('\n'):
            line_strip = line.strip()
            if line_strip:
                line_split = line_strip.split('#', 1)

                if len(line_split) > 1:
                    try:
                        procs[int(line_split[0])] = line_split[1].strip()
                    except ValueError:
                        continue

        return procs


async def map_processes() -> Dict[int, Tuple[str, str]]:
    pid_comm, pid_cmd = await asyncio.gather(_map_processes_by_attribute('c'), _map_processes_by_attribute('a'))

    if pid_comm and pid_cmd:
        res = {}
        for pid in {*pid_comm, *pid_cmd}:
            comm = pid_comm.get(pid)
            cmd = pid_cmd.get(pid)

            if comm and cmd:
                res[pid] = cmd, comm

        return res if res else None



import json
import sys
from logging import Logger
from typing import Optional

import aiohttp
from aiohttp import ServerDisconnectedError, ClientConnectorError

from guapow import __app_name__
from guapow.common import encryption
from guapow.common.config import OptimizerConfig
from guapow.common.dto import OptimizationRequest
from guapow.common.encoder import CustomJSONEncoder


async def send(request: OptimizationRequest, opt_config: OptimizerConfig, machine_id: Optional[str], logger: Logger):

    data = json.dumps(request.to_dict(), cls=CustomJSONEncoder)

    if machine_id:
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = encryption.encrypt(data, machine_id)
    else:
        headers = {'Content-Type': 'application/json'}

    try:
        logger.info("Sending request {}".format(request))
        async with aiohttp.ClientSession() as session:
            async with session.post(url=f'http://127.0.0.1:{opt_config.port}/',
                                    data=data,
                                    headers=headers) as response:
                if response.status in (200, 202):
                    logger.debug(f"Request successfully sent for pid '{request.pid}'")
                elif response.status == 401:
                    logger.warning(f"Unauthorized request. Optimizations will not be performed for pid '{request.pid}'")
                else:
                    logger.error(f"Unexpected response for the request (pid: {request.pid}, status: {response.status}. body: {await response.text()})")

    except ClientConnectorError:
        sys.stderr.write(f"[{__app_name__}] Request for pid '{request.pid}' could not reach the Optimizer service. It may not be running.\n")
    except ServerDisconnectedError:
        logger.warning(f"Request for pid '{request.pid}' reached the Optimizer service, but it did not respond")



import asyncio
import os
import re
import shutil
from abc import ABC, abstractmethod
from logging import Logger
from typing import Tuple, Optional, List, Set, Type, Dict

from guapow.common import system
from guapow.common.scripts import RunScripts
from guapow.runner.profile import RunnerProfile


class RunnerContext:

    def __init__(self, logger: Logger, processes_initialized: Optional[Set[int]], environment_variables: Optional[dict],
                 stopped_processes: Optional[Dict[str, str]]):
        self.logger = logger
        self.processes_initialized = processes_initialized
        self.environment_variables = environment_variables
        self.stopped_processes = stopped_processes

    def get_process_extra_arguments(self) -> dict:
        args = {}
        if self.environment_variables:
            args['env'] = {**os.environ, **self.environment_variables}

        return args


class RunnerTask(ABC):

    @abstractmethod
    def __init__(self, context: RunnerContext):
        pass

    @abstractmethod
    def is_available(self) -> Tuple[bool, Optional[str]]:
        pass

    @abstractmethod
    def should_run(self, profile: RunnerProfile) -> bool:
        pass

    @abstractmethod
    async def run(self, profile: RunnerProfile):
        pass


class AddDefinedEnvironmentVariables(RunnerTask):

    def __init__(self, context: RunnerContext):
        self.context = context
        self._log = context.logger

    def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    def should_run(self, profile: RunnerProfile) -> bool:
        return bool(profile.environment_variables)

    async def run(self, profile: RunnerProfile):
        vars_added = {}
        for var, val in profile.environment_variables.items():
            if var:
                final_val = str(val) if val is not None else ''
                self.context.environment_variables[var] = final_val
                vars_added[var] = final_val

        if vars_added:
            self._log.info(f"Environment variables added: {', '.join(f'{k} = {v}' for k,v in vars_added.items())}")


class RunPreLaunchScripts(RunnerTask):

    def __init__(self, context: RunnerContext):
        super(RunPreLaunchScripts, self).__init__(context)
        self._context = context
        self._task = RunScripts(name="pre launch", root_allowed=False, logger=context.logger)

    def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    def should_run(self, profile: RunnerProfile) -> bool:
        return bool(profile.before_scripts and profile.before_scripts.scripts)

    async def run(self, profile: RunnerProfile):
        pids = await self._task.run(scripts=[profile.before_scripts], user_id=os.getuid(), user_env={**os.environ})

        if pids:
            self._context.processes_initialized.update(pids)


class StopProcesses(RunnerTask):

    def __init__(self, context: RunnerContext):
        self._log = context.logger
        self._context = context

    def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    def should_run(self, profile: RunnerProfile) -> bool:
        return bool(profile.stop and profile.stop.processes)

    async def run(self, profile: RunnerProfile):
        stopped_processes = {}

        found = await system.find_pids_by_names(profile.stop.processes)
        not_stopped = set()

        if found:
            pid_cmds = await system.find_commands_by_pids({*found.values()})

            if not pid_cmds:
                self._log.warning(f"Could not retrieve commands of: {', '.join((str(p) for p in pid_cmds))}")

            _, kill_output = await system.async_syscall(f"kill -9 {' '.join((str(p) for p in found.values()))}")

            not_killed = re.compile(r'kill\s*:\s*\(?(\d+)\)?.+').findall(kill_output) if kill_output else []

            if not_killed:
                not_killed = {int(pid) for pid in not_killed}

            for comm, pid in found.items():
                if pid in not_killed:
                    not_stopped.add(comm)
                elif pid_cmds:
                    cmd = pid_cmds.get(pid)

                    if cmd:
                        stopped_processes[comm] = cmd

        if stopped_processes:
            self._log.info(f"Processes stopped: {', '.join(stopped_processes)}")

        if not_stopped:
            self._log.warning(f"Fail to stop processes: {', '.join(not_stopped)}")

        if len(stopped_processes) != len(profile.stop.processes):
            not_running = profile.stop.processes.difference(stopped_processes).difference(not_stopped)

            # check if the processes that were not running exist
            existent_not_running = {comm: None for comm in not_running if shutil.which(comm)}  # when the process command is sent to the backend with as 'None', it means it was not stopped. This behavior is essential to properly manage the requirements state.

            if existent_not_running:
                stopped_processes.update(existent_not_running)  # must be added even if not stopped so the optimizer can properly manage what should be relaunched
                self._log.warning(f"Some processes were not running and could not be stopped: {', '.join(existent_not_running)}")

        if stopped_processes:
            self._context.stopped_processes.update(stopped_processes)


class RunnerTaskManager:
    """
    Execute tasks before the process is called
    """

    def __init__(self, context: RunnerContext, actions: Optional[List[RunnerTask]] = None):
        self._context = context
        self._log = context.logger
        self._actions = actions
        self._tasks_order: Dict[Type[RunnerTask], int] = {
            AddDefinedEnvironmentVariables: 0,
            StopProcesses: 1,
            RunPreLaunchScripts: 2
        }

    def _sort(self, task: RunnerTask):
        return self._tasks_order.get(task.__class__, 99)

    def _check_availability(self):
        if self._actions is None:
            actions = []

            for cls in RunnerTask.__subclasses__():
                action = cls(self._context)
                available, warning_msg = action.is_available()

                if available:
                    actions.append(action)
                elif warning_msg:
                    self._log.warning(warning_msg)

            actions.sort(key=self._sort)
            self._actions = actions

    def get_available_tasks(self) -> List[RunnerTask]:
        return [*self._actions]

    async def run(self, profile: RunnerProfile) -> int:
        self._check_availability()

        if self._actions:
            res = await asyncio.gather(*(a.run(profile) for a in self._actions if a.should_run(profile)))
            return len(res)

        return 0



import os
import time
from datetime import datetime
from typing import Optional, Set, Dict


class OptimizationRequest:

    NOT_LOGGABLE_FIELDS = {'user_env', 'user_id'}

    def __init__(self, pid: Optional[int], command: Optional[str], user_name: Optional[str], profile: Optional[str] = None,
                 created_at: Optional[float] = time.time(), config: Optional[str] = None,
                 profile_config: Optional[str] = None, related_pids: Optional[Set[int]] = None,
                 user_env: Optional[dict] = None, stopped_processes: Optional[Dict[str, Optional[str]]] = None,
                 relaunch_stopped_processes: Optional[bool] = None):
        self.pid = pid
        self.command = command
        self.profile = profile
        self.user_name = user_name
        self.created_at = created_at
        self.config = config
        self.profile_config = profile_config
        self.related_pids = related_pids if related_pids is None or isinstance(related_pids, set) else {*related_pids}  # processes that were initialized with/for the root process
        self.user_env = user_env  # only informed to scripts to be executed at the user level
        self.stopped_processes = stopped_processes
        self.relaunch_stopped_processes = relaunch_stopped_processes
        self.user_id: Optional[int] = None

    def is_valid(self) -> bool:
        return self.pid is not None and self.pid >= 0 and \
               bool(self.command) and \
               bool(self.user_name) and self.user_id is not None

    def to_dict(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def has_full_configuration(self) -> bool:
        return bool(self.config)

    @property
    def is_self_request(self) -> bool:
        return self.pid is None and self.command is None and self.user_name is None

    @classmethod
    def new_from_config(cls, config: str, **kwargs) -> "OptimizationRequest":
        if kwargs and 'profile_config' in kwargs:
            del kwargs['profile_config']

        return cls(config=config, **kwargs)

    @classmethod
    def new_from_profile(cls, **kwargs) -> "OptimizationRequest":
        if kwargs and 'config' in kwargs:
            del kwargs['config']

        return cls(**kwargs)

    def prepare(self):
        """sets up required/important properties"""

        if self.user_env is None:
            self.user_env = {}

        if 'DISPLAY' not in self.user_env:
            self.user_env['DISPLAY'] = os.getenv('DISPLAY', ':0')

    def __repr__(self) -> str:
        return ', '.join(['{}: {}'.format(k, str(datetime.fromtimestamp(v)) if k == 'created_at' else v) for k, v in self.__dict__.items() if v is not None and k not in self.NOT_LOGGABLE_FIELDS]).replace('\n', ' ')

    def __eq__(self, other):
        if isinstance(other, OptimizationRequest):
            return self.__dict__ == other.__dict__

    @classmethod
    def self_request(cls) -> "OptimizationRequest":
        return cls(pid=None, command=None, user_name=None, created_at=time.time())



from typing import Optional, Set, Dict, Tuple

from guapow import __app_name__
from guapow.common.model import FileModel
from guapow.common.users import is_root_user


def get_default_profile_name() -> str:
    return 'default'


def get_root_profile_path(name: str) -> str:
    return f'/etc/{__app_name__}/{name}.profile'


def get_user_profile_path(name: str, user_name: str) -> str:
    return f'/home/{user_name}/.config/{__app_name__}/{name}.profile'


def get_profile_dir(user_id: int, user_name: str) -> str:
    return f'/etc/{__app_name__}' if is_root_user(user_id) else f'/home/{user_name}/.config/{__app_name__}'


def get_possible_profile_paths_by_priority(name: str, user_id: Optional[int], user_name: Optional[str]) -> Tuple[str, Optional[str]]:  # using tuple instead of list to reduce memory usage
    is_root = is_root_user(user_id)

    if not is_root and user_name:
        return get_user_profile_path(name, user_name), get_root_profile_path(name)
    else:
        return get_root_profile_path(name), None


class StopProcessSettings(FileModel):

    def __init__(self, node_name: str, processes: Optional[Set[str]], relaunch: Optional[bool] = None):
        self.processes = processes
        self.relaunch = relaunch
        self._mapping = {f"{node_name}{f'.{p}' if p else ''}": v for p, v in
                         {'': ('processes', Set[str], None), 'relaunch': ('relaunch', bool, True)}.items()}

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self._mapping

    def is_valid(self) -> bool:
        return bool(self.processes)

    def get_file_root_node_name(self) -> Optional[str]:
        pass



import re
from enum import Enum
from logging import Logger
from re import Pattern
from typing import Optional, Set, Tuple, Dict, Collection

from guapow.common import steam
from guapow.common.util import map_only_any_regex, has_any_regex

RE_PYTHON_REGEX = re.compile(r'^r:(.+)$')


class RegexType(Enum):
    CMD = 0
    COMM = 1


class RegexMapper:

    BUILTIN_RE = {'__steam__': (steam.RE_STEAM_CMD, RegexType.CMD)}

    def __init__(self, cache: bool, logger: Logger):
        self._pattern_cache: Optional[dict] = None
        self._string_no_pattern_cache: Optional[Set[str]] = None

        if cache:
            self._pattern_cache, self._string_no_pattern_cache = {}, set()

        self._log = logger

    def get_cached_pattern(self, str_pattern: str) -> Optional[Pattern]:
        if self._pattern_cache is not None:
            return self._pattern_cache.get(str_pattern)

    def is_cached_as_no_pattern(self, string: str) -> bool:
        return string in self._string_no_pattern_cache if self._string_no_pattern_cache is not None else False

    def map(self, string: str) -> Optional[Tuple[Pattern, RegexType]]:
        builtin_pattern = self.BUILTIN_RE.get(string)

        if builtin_pattern:
            if builtin_pattern[1] == RegexType.CMD:
                return builtin_pattern[0], RegexType.CMD

            if builtin_pattern[1] == RegexType.COMM:
                return builtin_pattern[0], RegexType.COMM

            self._log.error(f"Unknown type of built-in pattern '{string}'. It will be ignored.")
            return

        if self.is_cached_as_no_pattern(string):
            return

        pattern = self.get_cached_pattern(string)
        cached = bool(pattern)

        if not cached:
            python_regex = RE_PYTHON_REGEX.findall(string)

            if python_regex:
                try:
                    pattern = re.compile('^{}$'.format(python_regex[0]))
                except re.error:
                    self._log.warning(f'Invalid Python regex mapping: {string}')
                    return

            elif has_any_regex(string):
                pattern = map_only_any_regex(string)
            else:
                pattern = None

        if pattern:
            type_ = RegexType.CMD if pattern.pattern[1] == '/' else RegexType.COMM

            if not cached and self._pattern_cache is not None:
                self._pattern_cache[string] = pattern

            return pattern, type_

        elif self._string_no_pattern_cache is not None:
            self._string_no_pattern_cache.add(string)

    def map_for_profiles(self, mapping: Dict[str, str]) -> Optional[Tuple[Dict[Pattern, str], Dict[Pattern, str]]]:
        """
        return: a tuple with two dictionaries: first with cmd patterns and second with comm patterns.
        """
        if mapping:
            cmd, comm = {}, {}

            for string, prof in mapping.items():
                string_mapping = self.map(string)

                if string_mapping:
                    pattern, type_ = string_mapping

                    if type_ == RegexType.CMD:
                        cmd[pattern] = prof
                    elif type_ == RegexType.COMM:
                        comm[pattern] = prof
                    elif type_:
                        self._log.error(f"Could not assign pattern ({string}) to profile ({prof}). "
                                        f"Unsupported {RegexType.__class__.__name__} ({type_})")
                    else:
                        self._log.error(f"No {RegexType.__class__.__name__} returned when mapping pattern ({string})"
                                        f" to profile ({prof})")

            return cmd, comm

    def map_collection(self, strings: Collection[str]) -> Optional[Dict[RegexType, Set[Pattern]]]:
        if strings:
            res: Dict[RegexType, Set[Pattern]] = {type_: None for type_ in RegexType}
            for string in strings:
                string_mapping = self.map(string)
                if string_mapping:
                    pattern, type_ = string_mapping

                    if type_:
                        patterns = res.get(type_)

                        if patterns is None:
                            patterns = set()
                            res[type_] = patterns

                        patterns.add(pattern)
                    else:
                        self._log.error(f"No {RegexType.__class__.__name__} returned when mapping pattern ({string})")

            return res



import asyncio
import multiprocessing
import os.path
import traceback
from asyncio import Lock
from logging import Logger
from typing import Optional, Set, Dict, Tuple

import aiofiles

GOVERNOR_FILE_PATTERN = '/sys/devices/system/cpu/cpu{}/cpufreq/scaling_governor'
GOVERNOR_PERFORMANCE = 'performance'


def get_cpu_count() -> int:
    try:
        return multiprocessing.cpu_count()
    except:
        return 0


class CPUFrequencyManager:

    def __init__(self, logger: Logger, cpu_count: int, governor_file_pattern: Optional[str] = GOVERNOR_FILE_PATTERN):
        self._cached_governors = {}
        self._log = logger
        self._governor_file_pattern = governor_file_pattern
        self._cpu_count = cpu_count
        self._lock = Lock()

    def lock(self):
        return self._lock

    def save_governors(self, governor_cpus: Dict[str, Set[int]]):
        if self._cached_governors is None:
            self._cached_governors = {}

        for gov, cpus in governor_cpus.items():
            for cpu in cpus:
                self._cached_governors[cpu] = gov

    def get_saved_governors(self) -> Optional[Dict[str, Set[int]]]:
        governors = None
        if self._cached_governors:
            governors = {}
            for cpu, gov in self._cached_governors.items():
                cpus = governors.get(gov, set())
                governors[gov] = cpus
                cpus.add(cpu)

        return governors

    async def map_current_governors(self) -> Dict[str, Set[int]]:
        governors = {}
        if self._cpu_count > 0:
            for cpu in range(self._cpu_count):
                governor_path = self._governor_file_pattern.format(cpu)
                try:
                    async with aiofiles.open(governor_path) as f:
                        gov = await f.read()
                except FileNotFoundError:
                    self._log.warning(f"Could not read governor for CPU '{cpu}'. File '{governor_path}' not found")
                    continue

                gov = gov.strip()
                cpus = governors.get(gov, set())
                cpus.add(cpu)
                governors[gov] = cpus

        return governors

    async def _write_governor(self, idx: int, governor: str) -> bool:
        try:
            async with aiofiles.open(self._governor_file_pattern.format(idx), 'w+') as f:
                await f.write(governor)

            return True
        except OSError:
            return False

    async def _set_governor(self, idx: int, governor: str, changed: Set[int], not_changed: Set[int]):
        if await self._write_governor(idx, governor):
            changed.add(idx)
        else:
            not_changed.add(idx)

    async def change_governor(self, governor: str, cpu_idxs: Optional[Set[int]] = None) -> Set[int]:
        if self._cpu_count == 0:
            return set()

        changed, not_changed = set(), set()

        await asyncio.gather(*(self._set_governor(i, governor, changed, not_changed) for i in (cpu_idxs if cpu_idxs else range(self._cpu_count))))

        if not_changed:
            self._log.warning(f"Could not change CPUs [{','.join((str(i) for i in not_changed))}] frequency governor to '{governor}'")

        if changed:
            self._log.info(f"CPUs [{','.join((str(i) for i in changed))}] frequency governor changed to '{governor}'")

        return changed


class CPUEnergyPolicyManager:

    LEVEL_PERFORMANCE = 0

    def __init__(self, cpu_count: int, logger: Logger,
                 file_pattern: str = '/sys/devices/system/cpu/cpu{idx}/power/energy_perf_bias'):
        self._cpus = cpu_count
        self._log = logger
        self._lock = Lock()
        self._state_cache: Dict[int, int] = dict()
        self._file_pattern = file_pattern

    def lock(self) -> Lock:
        return self._lock

    def can_work(self) -> Tuple[bool, Optional[str]]:
        if not self._cpus or self._cpus < 0:
            return False, 'It will not be possible to change the CPU energy policy level: no CPU detected'

        file_path = self._file_pattern.format(idx='0')
        if not os.path.exists(file_path):
            return False, f"It will not be possible to change the CPU energy policy level: " \
                          f"file '{file_path}' not found"

        return True, None

    async def _read_cpu_state(self, idx: int) -> Optional[int]:
        file_path = self._file_pattern.format(idx=idx)
        cpu_state = None

        try:
            async with aiofiles.open(file_path) as f:
                cpu_state = await f.read()
        except:
            err_stack = traceback.format_exc().replace('\n', ' ')
            self._log.error(f"[{self.__class__.__name__}] Could not read file '{file_path}': {err_stack}")

        if cpu_state is not None:
            try:
                return int(cpu_state)
            except ValueError:
                self._log.error(f"[{self.__class__.__name__}] Could not cast CPU energy policy level ({cpu_state}) "
                                f"to int. File: {file_path}")

    async def _write_cpu_state(self, idx: int, state: int) -> bool:
        file_path = self._file_pattern.format(idx=idx)

        try:
            async with aiofiles.open(file_path, 'w+') as f:
                await f.write(str(state))

            return True
        except:
            err_stack = traceback.format_exc().replace('\n', ' ')
            self._log.error(f"[{self.__class__.__name__}] Could not write '{state}' to file '{file_path}': {err_stack}")
            return False

    async def _fill_cpu_state_change(self, idx: int, state: int, output: Dict[int, bool]):
        output[idx] = await self._write_cpu_state(idx, state)

    async def _fill_cpu_state(self, idx: int, output: Dict[int, int]):
        cpu_state = await self._read_cpu_state(idx)

        if isinstance(cpu_state, int):
            output[idx] = cpu_state

    async def map_current_state(self) -> Optional[Dict[int, int]]:
        if self._cpus > 0:
            res = {}
            await asyncio.gather(*[self._fill_cpu_state(idx, res) for idx in range(self._cpus)])
            return res if res else None

    async def change_states(self, cpu_states: Dict[int, int]) -> Dict[int, bool]:
        if self._cpus > 0 and cpu_states:
            res = {}
            await asyncio.gather(*[self._fill_cpu_state_change(idx, state, res) for idx, state in cpu_states.items()])
            return res

    def save_state(self, cpu_states: Dict[int, int]):
        if cpu_states:
            for idx, state in cpu_states.items():
                if idx not in self._state_cache:
                    self._state_cache[idx] = state

    @property
    def saved_state(self) -> Dict[int, int]:
        return {**self._state_cache}

    def clear_state(self, *keys: int):
        if keys:
            for k in keys:
                if k in self._state_cache:
                    del self._state_cache[k]
        else:
            self._state_cache.clear()



import asyncio
import os
import re
import shutil
import traceback
from abc import ABC, abstractmethod
from asyncio import Lock
from copy import deepcopy
from glob import glob
from logging import Logger, ERROR, DEBUG, WARNING, INFO
from re import Pattern
from typing import Optional, Tuple, Set, Dict, List, Type, AsyncIterator, Any

import aiofiles

from guapow.common import system
from guapow.common.model import CustomEnum


class UnknownGPUDriver(Exception):

    def __init__(self, gpu_idx: int):
        self.gpu_idx = gpu_idx


class GPUDriver(ABC):

    @abstractmethod
    def __init__(self, cache: bool, logger: Logger):
        self._logger = logger
        self._lock = Lock()
        self._cache_lock = Lock() if cache else None
        self._gpus: Optional[Set[str]] = None
        self._cached = False

    def lock(self) -> Lock:
        return self._lock

    async def get_cached_gpus(self) -> Optional[Set[str]]:
        if self._cache_lock is None:
            return await self.get_gpus()

        async with self._cache_lock:
            if not self._cached:
                self._gpus = await self.get_gpus()
                self._cached = True

        return self._gpus

    @abstractmethod
    async def get_gpus(self) -> Optional[Set[str]]:
        pass

    @classmethod
    @abstractmethod
    def get_vendor_name(cls) -> str:
        pass

    @abstractmethod
    async def set_power_mode(self, ids_modes: Dict[str, Any], user_environment: Optional[Dict[str, str]]) \
            -> Dict[str, bool]:
        pass

    @abstractmethod
    async def get_power_mode(self, gpu_ids: Set[str], user_environment: Optional[Dict[str, str]]) \
            -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def can_work(self) -> Tuple[bool, Optional[str]]:
        pass

    @abstractmethod
    def get_default_mode(self) -> Any:
        pass

    @abstractmethod
    def get_performance_mode(self) -> Any:
        pass

    def _log(self, msg: str, level: int = INFO):
        final_msg = f"{self.get_vendor_name()}: {msg}"
        if level == INFO:
            self._logger.info(final_msg)
        elif level == DEBUG:
            self._logger.debug(final_msg)
        elif level == ERROR:
            self._logger.error(final_msg)
        elif level == WARNING:
            self._logger.warning(final_msg)
        else:
            self._logger.info(final_msg)


class NvidiaPowerMode(CustomEnum):
    ON_DEMAND = 0
    PERFORMANCE = 1
    AUTO = 2


class NvidiaGPUDriver(GPUDriver):

    def __init__(self, cache: bool, logger: Logger):
        super(NvidiaGPUDriver, self).__init__(cache, logger)
        self._re_set_power: Optional[Pattern] = None
        self._re_get_power: Optional[Pattern] = None

    def _get_re_set_power(self) -> Pattern:
        if self._re_set_power is None:
            self._re_set_power = re.compile(r'\[gpu:(\d+)].+(\d)\.?')

        return self._re_set_power

    def _get_re_get_power(self) -> Pattern:
        if self._re_get_power is None:
            self._re_get_power = re.compile(r"Attribute\s+.+\[gpu:(\d+)].+:\s+(\d)")

        return self._re_get_power

    @classmethod
    def get_vendor_name(cls) -> str:
        return 'Nvidia'

    async def get_gpus(self) -> Optional[Set[str]]:
        exitcode, output = await system.async_syscall('nvidia-smi --query-gpu=index --format=csv,noheader')

        gpus = set()
        if exitcode == 0:
            for line in output.split('\n'):
                if line:
                    gpu_idx = line.strip()
                    if gpu_idx:
                        gpus.add(gpu_idx)

        return gpus

    def _map_env_vars(self, vars: Optional[Dict[str, str]]) -> Dict[str, str]:
        env = vars if vars is not None else {}
        env['LANG'] = 'en_US.UTF-8'
        return env

    async def set_power_mode(self, ids_modes: Dict[str, NvidiaPowerMode],
                             user_environment: Optional[Dict[str, str]] = None) -> Dict[str, bool]:
        params = ' '.join((f'-a [gpu:{i}]/GpuPowerMizerMode={m.value}' for i, m in ids_modes.items()))
        cmd = f'nvidia-settings {params}'

        log_str = {', '.join((f'{i}={ids_modes[i].value}' for i in ids_modes))}
        self._log(f"changing GPUs power mode ({log_str}): {cmd}")
        _, output = await system.async_syscall(cmd, custom_env=self._map_env_vars(user_environment))

        if output:
            changed_gpus = {*self._get_re_set_power().findall(output)}

            if changed_gpus:
                try:
                    return {id_: int(mode) == ids_modes[id_].value for id_, mode in changed_gpus if id_ in ids_modes}
                except ValueError:
                    self._log(f"error while parsing changing modes response: {output}", ERROR)

        err_msg = output.replace('\n', ' ') if output else ''
        self._log(f"could not determine the changing modes response: {err_msg}", ERROR)
        return {i: False for i in ids_modes}

    async def get_power_mode(self, gpu_ids: Set[str], user_environment: Optional[Dict[str, str]] = None) \
            -> Optional[Dict[str, NvidiaPowerMode]]:
        if gpu_ids:
            gpus_query = ' '.join((f'-q [gpu:{id_}]/GpuPowerMizerMode' for id_ in gpu_ids))
            cmd = f"nvidia-settings {gpus_query}"
            code, output = await system.async_syscall(cmd, custom_env=self._map_env_vars(user_environment))

            if code == 0:
                if not output:
                    self._log(f"could not detect GPUs power mode ({cmd}). No output returned", WARNING)
                else:
                    modes = self._get_re_get_power().findall(output)

                    if modes:
                        try:
                            return {id_: NvidiaPowerMode.from_value(int(mode)) for id_, mode in modes if id_ in gpu_ids}
                        except ValueError:
                            self._log(f"error when parsing power modes: {modes}", ERROR)

                    self._log(f"could not detect GPUs power mode ({cmd}). No modes found in output: {output}", ERROR)
            else:
                output_str = '. Output: {}'.format(output.replace('\n', ' ')) if output else ''
                self._log(f"could not detect GPUs power mode ({cmd}){output_str}", ERROR)

    def can_work(self) -> Tuple[bool, Optional[str]]:
        if not shutil.which('nvidia-settings'):
            return False, "'nvidia-settings' is not installed"

        if not shutil.which('nvidia-smi'):
            return False, "'nvidia-smi' is not installed"

        return True, None

    def get_default_mode(self) -> NvidiaPowerMode:
        return NvidiaPowerMode.AUTO

    def get_performance_mode(self) -> NvidiaPowerMode:
        return NvidiaPowerMode.PERFORMANCE


class AMDGPUDriver(GPUDriver):

    PERFORMANCE_FILE = "power_dpm_force_performance_level"
    PROFILE_FILE = "pp_power_profile_mode"
    VENDOR = "AMD"

    def __init__(self, cache: bool, logger: Logger, gpus_path: str = "/sys/class/drm/card{id}/device"):
        super(AMDGPUDriver, self).__init__(cache, logger)
        self._gpus_path = gpus_path
        self._re_power_mode: Optional[Pattern] = None
        self._re_extract_id: Optional[Pattern] = None

    @classmethod
    def get_vendor_name(cls) -> str:
        return cls.VENDOR

    def can_work(self) -> Tuple[bool, Optional[str]]:
        return True, None

    @property
    def re_power_mode(self) -> Pattern:
        if not self._re_power_mode:
            self._re_power_mode = re.compile(r'^\w+\*:?$')

        return self._re_power_mode

    @property
    def re_extract_id(self) -> Pattern:
        if not self._re_extract_id:
            self._re_extract_id = re.compile(self._gpus_path.replace('{id}', r'(\d+)'))

        return self._re_extract_id

    def extract_gpu_id(self, gpu_path: str) -> Optional[int]:
        try:
            return self.re_extract_id.findall(gpu_path)[0]
        except IndexError:
            self._log(f"Could not extract GPU id from path: {gpu_path}", ERROR)

    async def get_gpus(self) -> Optional[Set[str]]:
        required_files = {self.PERFORMANCE_FILE: set(), self.PROFILE_FILE: set()}

        for gpu_file_path in glob(f"{self._gpus_path.format(id='*')}/*"):
            gpu_file = os.path.basename(gpu_file_path)
            if gpu_file in required_files:
                if not os.access(gpu_file_path, mode=os.W_OK):
                    id_ = self.extract_gpu_id(gpu_file_path)
                    self._log(f"Writing is not allowed for file '{gpu_file_path}. It will not be possible to set "
                              f"the GPU ({id_}) to performance mode", WARNING)
                else:
                    required_files[gpu_file].add(os.path.dirname(gpu_file_path))

        all_gpu_dirs = {gpu_dir for paths in required_files.values() for gpu_dir in paths}

        if all_gpu_dirs:
            gpus = set()
            for gpu_dir in all_gpu_dirs:
                missing_files = set()
                for file, gpu_file_dirs in required_files.items():
                    if gpu_dir not in gpu_file_dirs:
                        missing_files.add(file)

                if missing_files:
                    self._log(f"not all required files are accessible for mounted GPU in '{gpu_dir}' "
                              f"(missing: {', '.join(sorted(missing_files))})", WARNING)
                else:
                    self._log(f"all required files are accessible for GPU mounted in '{gpu_dir}'", DEBUG)
                    gpu_id = self.extract_gpu_id(gpu_dir)

                    if gpu_id is not None:
                        gpus.add(gpu_id)

            return gpus if gpus else None
        else:
            self._log("no mounted GPU directories", DEBUG)

    async def _read_file(self, file_path: str) -> Optional[str]:
        try:
            async with aiofiles.open(file_path) as f:
                return await f.read()
        except:
            err_stack = traceback.format_exc().replace('\n', ' ')
            self._log(f"Could not read file '{file_path}': {err_stack}", ERROR)

    def _map_power_profile_output(self, output: str, file_path: str) -> Optional[str]:
        if output is not None:
            for raw_line in output.split('\n'):
                if raw_line.startswith(' '):
                    line = raw_line.strip().split(' ')

                    if len(line) > 1 and line[0].isdigit() and self.re_power_mode.match(line[-1]):
                        return line[0].strip()

            content_log = output.replace('\n', ' ')
            self._log(f"could not map power profile from {file_path}. Content: {content_log}", WARNING)

    async def _fill_power_mode(self, gpu_id: str, gpu_modes: Dict[str, str]):
        gpu_dir = self._gpus_path.format(id=gpu_id)
        performance_level_file = f"{gpu_dir}/{self.PERFORMANCE_FILE}"
        performance_level = (await self._read_file(performance_level_file)).strip()
        self._log(f"GPU file ({performance_level_file}): {performance_level}", DEBUG)

        if not performance_level:
            return

        power_profile_file = f"{gpu_dir}/{self.PROFILE_FILE}"
        power_profile = self._map_power_profile_output(await self._read_file(power_profile_file), power_profile_file)
        self._log(f"GPU file ({power_profile_file}): {power_profile}", DEBUG)

        if not power_profile:
            return

        gpu_modes[gpu_id] = f"{performance_level}:{power_profile}"

    async def get_power_mode(self, gpu_ids: Set[str], user_environment: Optional[Dict[str, str]] = None) \
            -> Optional[Dict[str, str]]:
        if gpu_ids:
            res = {}
            await asyncio.gather(*tuple(self._fill_power_mode(id_, res) for id_ in gpu_ids))
            return res if res else None

    async def _write_to_file(self, file_path: str, content: str) -> bool:
        try:
            async with aiofiles.open(file_path, 'w+') as f:
                await f.write(content)
            return True
        except:
            err_stack = traceback.format_exc().replace('\n', ' ')
            self._log(f"could not write '{content}' to file '{file_path}': {err_stack}", ERROR)
            return False

    async def _fill_write_result(self, file_path: str, content: str, id_: str, output: Dict[str, List[bool]]):
        output[id_].append(await self._write_to_file(file_path, content))

    async def set_power_mode(self, ids_modes: Dict[str, str],
                             user_environment: Optional[Dict[str, str]] = None) -> Dict[str, bool]:
        res = {}
        if ids_modes:
            coros, writes = [], dict()
            for id_, mode_str in ids_modes.items():
                mode = mode_str.split(':')

                if len(mode) == 2:
                    gpu_dir = self._gpus_path.format(id=id_)
                    self._log(f"changing GPU ({id_}) operation mode (performance: {mode[0]}, profile: {mode[1]})")
                    writes[id_] = list()
                    coros.append(self._fill_write_result(f'{gpu_dir}/{self.PERFORMANCE_FILE}', mode[0], id_, writes))
                    coros.append(self._fill_write_result(f'{gpu_dir}/{self.PROFILE_FILE}', mode[1], id_, writes))
                else:
                    self._log(f"could not change GPU ({id_}) operation mode: unexpected mode format '{mode_str}' "
                              f"(expected: 'performance_level:power_profile'", ERROR)

            await asyncio.gather(*coros)

            for id_ in ids_modes:
                gpu_writes = writes.get(id_)
                res[id_] = gpu_writes and all(gpu_writes)

        return res

    def get_default_mode(self) -> str:
        return 'auto:3'

    def get_performance_mode(self) -> str:
        return 'manual:5'


class GPUState:

    def __init__(self, id_: str, driver_class: Type, power_mode: Any):
        self.id = id_
        self.driver_class = driver_class
        self.power_mode = power_mode

    def __eq__(self, other):
        if isinstance(other, GPUState):
            return self.driver_class == other.driver_class and self.id == other.id

    def __hash__(self):
        return hash(self.driver_class) + hash(self.id)

    def __repr__(self):
        attrs = self.__dict__.items()
        attr_str = ', '.join(f'{p}={v.__name__ if isinstance(v, type) else v}' for p, v in sorted(attrs) if v)
        return f'{self.__class__.__name__} ({attr_str})'


class GPUManager:

    LOG_CACHE_KEY__WORK = 0
    LOG_CACHE_KEY__AVAILABLE = 1

    def __init__(self, logger: Logger, drivers: Optional[Tuple[GPUDriver]] = None, cache_gpus: bool = False):
        self._log = logger
        self._drivers = drivers
        self._drivers_lock = Lock()
        self._cache_gpus = cache_gpus
        self._gpu_state_cache: Dict[Type[GPUDriver], Dict[str, Any]] = {}
        self._gpu_state_cache_lock = Lock()
        self._log_cache: Dict[Type[GPUDriver], Dict[int, object]] = {}  # to avoid repetitive logs
        self._working_drivers_cache: Optional[Tuple[GPUDriver]] = None  # only when 'cache_gpus'
        self._working_drivers_cache_lock = Lock()

    def is_cache_enabled(self) -> bool:
        return self._cache_gpus

    def _get_driver_log_cache(self, cls: Type[GPUDriver]) -> Dict[int, object]:
        driver_cache = self._log_cache.get(cls)

        if driver_cache is None:
            driver_cache = {}
            self._log_cache[cls] = driver_cache

        return driver_cache

    async def _can_driver_work(self, driver: GPUDriver) -> bool:
        can_work, reason = driver.can_work()

        driver_cache = self._get_driver_log_cache(driver.__class__)

        if can_work:
            driver_cache[self.LOG_CACHE_KEY__WORK] = False
        else:
            logged = driver_cache.get(self.LOG_CACHE_KEY__WORK)

            if not logged:
                self._log.warning(f"{driver.get_vendor_name()} GPUs cannot be managed: "
                                  f"{reason if reason else 'unknown reason'}")
                driver_cache[self.LOG_CACHE_KEY__WORK] = True

        return can_work

    async def _get_driver_gpus(self, driver: GPUDriver) -> Optional[Set[str]]:
        gpus = await driver.get_cached_gpus()

        driver_cache = self._get_driver_log_cache(driver.__class__)

        cached_gpus = driver_cache.get(self.LOG_CACHE_KEY__AVAILABLE)

        if gpus != cached_gpus:
            gpu_ids = f" (ids={', '.join((str(i) for i in sorted(gpus)))})" if gpus else ''
            self._log.debug(f'[{driver.get_vendor_name()}] GPUs available: {len(gpus)}{gpu_ids}')
            driver_cache[self.LOG_CACHE_KEY__AVAILABLE] = gpus

        return gpus

    async def _map_driver_if_gpus(self, driver: GPUDriver) -> Optional[Tuple[GPUDriver, Optional[Set[str]]]]:
        if await self._can_driver_work(driver):
            gpus = await self._get_driver_gpus(driver)

            if gpus:
                return driver, gpus

    async def _map_drivers_and_gpus(self) -> AsyncIterator[Tuple[GPUDriver, Set[str]]]:
        for task in asyncio.as_completed(tuple(self._map_driver_if_gpus(driver) for driver in self._drivers)):
            driver_gpus = await task

            if driver_gpus:
                yield driver_gpus

    async def map_working_drivers_and_gpus(self) -> AsyncIterator[Tuple[GPUDriver, Set[str]]]:
        async with self._drivers_lock:
            if self._drivers is None:
                driver_types = GPUDriver.__subclasses__()
                self._drivers = tuple(cls(self._cache_gpus, self._log) for cls in driver_types if cls != self.__class__)

        if self._drivers:
            if self._cache_gpus:
                async with self._working_drivers_cache_lock:
                    if self._working_drivers_cache is not None:
                        for driver in self._working_drivers_cache:
                            yield driver, await self._get_driver_gpus(driver)
                    else:
                        working_drivers = []

                        async for driver, gpus in self._map_drivers_and_gpus():
                            yield driver, gpus
                            working_drivers.append(driver)

                        self._working_drivers_cache = tuple(working_drivers)
            else:
                async for driver, gpus in self._map_drivers_and_gpus():
                    yield driver, gpus
        else:
            self._log.error("No GPU driver instances available")

    async def activate_performance(self, user_environment: Optional[Dict[str, str]] = None,
                                   target_gpu_ids: Optional[Set[str]] = None) \
            -> Optional[Dict[Type[GPUDriver], Set[GPUState]]]:
        """

        Args:
            user_environment: user environment variables
            target_gpu_ids: the target GPU ids to enter in performance mode. If None, all available GPUs will be considered.

        Returns: the GPUs previous states

        """

        res = {}
        async for driver, gpus in self.map_working_drivers_and_gpus():
            if not gpus:
                continue

            target_gpus = gpus.intersection(target_gpu_ids) if target_gpu_ids else gpus

            if not target_gpus:
                self._log.debug(f"[{driver.get_vendor_name()}] No valid target GPUs available "
                                f"for performance mode (valid: {', '.join(sorted(gpus))})")
                continue

            async with driver.lock():
                if target_gpu_ids and gpus != target_gpu_ids:
                    self._log.debug(f"[{driver.get_vendor_name()}] Target GPU ids for performance mode: "
                                    f"{', '.join(sorted(target_gpus))}")

                gpu_modes = await driver.get_power_mode(target_gpus, user_environment)
                if gpu_modes:
                    performance_mode = driver.get_performance_mode()
                    async with self._gpu_state_cache_lock:
                        cached_states = self._gpu_state_cache.get(driver.__class__, {})
                        self._gpu_state_cache[driver.__class__] = cached_states

                        driver_res, not_in_performance = set(), set()
                        for gpu, mode in gpu_modes.items():
                            if performance_mode != mode:
                                cached_states[gpu] = mode
                                driver_res.add(GPUState(gpu, driver.__class__, mode))
                                not_in_performance.add(gpu)
                            else:
                                old_state = cached_states.get(gpu)

                                if old_state:
                                    driver_res.add(GPUState(gpu, driver.__class__, old_state))

                    if not_in_performance:
                        gpus_changed = await driver.set_power_mode({g: performance_mode for g in not_in_performance},
                                                                   user_environment)

                        not_changed = {gpu for gpu, changed in gpus_changed.items() if not changed}

                        if not_changed:
                            self._log.error(f"[{driver.get_vendor_name()}] could not change power mode of GPUs: "
                                            f"{', '.join(sorted(not_changed))}")

                    res[driver.__class__] = driver_res

        return res

    def get_drivers(self) -> Optional[Tuple[GPUDriver]]:
        return tuple(self._drivers) if self._drivers is not None else None

    def get_cached_working_drivers(self) -> Optional[Tuple[GPUDriver]]:
        if self._working_drivers_cache:
            return tuple(self._working_drivers_cache)

    def get_gpu_state_cache_view(self) -> Dict[Type[GPUDriver], Dict[str, Any]]:
        return deepcopy(self._gpu_state_cache)


def get_driver_by_vendor(vendor: str) -> Optional[Type[GPUDriver]]:
    if vendor:
        vendor_norm = vendor.strip().lower()

        for cls_ in GPUDriver.__subclasses__():
            if cls_ != GPUManager and cls_.get_vendor_name().strip().lower() == vendor_norm:
                return cls_



import asyncio
import os
import subprocess
import traceback
from io import StringIO
from multiprocessing import Manager, Process
from multiprocessing.managers import DictProxy
from re import Pattern
from typing import Optional, Tuple, Dict, List, Set, Callable, TypeVar, Collection

BAD_USER_ENV_VARS = {'LD_PRELOAD'}
T = TypeVar('T')


def syscall(cmd: str, shell: bool = True, cwd: Optional[str] = None, custom_env: Optional[dict] = None, stdin: bool = False, return_output: bool = True) -> Tuple[int, Optional[str]]:
    params = {
        'args': cmd.split(' ') if not shell else [cmd],
        'stdout': subprocess.PIPE if return_output else subprocess.DEVNULL,
        'stderr': subprocess.STDOUT if return_output else subprocess.DEVNULL,
        'shell': shell
    }

    if not stdin:
        params['stdin'] = subprocess.DEVNULL

    if cwd is not None:
        params['cwd'] = cwd

    if custom_env is not None:
        params['env'] = custom_env

    try:
        p = subprocess.Popen(**params)
    except:
        traceback.print_exc()
        return False, None

    str_output = StringIO('') if return_output else None

    if str_output:
        for stream in p.stdout:
            string = stream.decode()

            if return_output:
                str_output.write(string)

    p.wait()

    if str_output and return_output:
        str_output.seek(0)
        return p.returncode, str_output.read()
    else:
        return p.returncode, None


async def async_syscall(cmd: str, shell: bool = True, cwd: Optional[str] = None, custom_env: Optional[Dict[str, str]] = None, stdin: bool = False, return_output: bool = True, wait: bool = True) -> Tuple[int, Optional[str]]:
    params = {
        'cmd': cmd,
        'stdout': subprocess.PIPE if return_output else subprocess.DEVNULL,
        'stderr': subprocess.STDOUT if return_output else subprocess.DEVNULL,
    }

    if not stdin:
        params['stdin'] = subprocess.DEVNULL

    if cwd is not None:
        params['cwd'] = cwd

    if custom_env is not None:
        params['env'] = custom_env

    try:
        if shell:
            p = await asyncio.create_subprocess_shell(**params)
        else:
            p = await asyncio.create_subprocess_exec(**params)
    except:
        traceback.print_exc()
        return False, None

    output = None

    if return_output:
        string = StringIO()

        try:
            async for stream in p.stdout:
                if stream:
                    string.write(stream.decode())
        except ValueError:
            return 1, None

        string.seek(0)
        output = string.read()

    if wait:
        await p.wait()

    return p.returncode, (output if output is not None else None)


async def match_syscall(cmd: str, match: Callable[[str], Optional[T]]) -> Optional[T]:
    p = await asyncio.create_subprocess_shell(cmd=cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL)

    try:
        async for stream in p.stdout:
            if stream:
                res = match(stream.decode())

                if res is not None:
                    return res

    except ValueError:
        return


async def find_process_by_name(name_pattern: Pattern, last_match: bool = False) -> Optional[Tuple[int, str]]:
    def _match(line: str) -> Optional[Tuple[int, str]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                name = line_split[1].strip()
                if name_pattern.match(name):
                    try:
                        return int(line_split[0]), name
                    except ValueError:
                        return

    return await match_syscall(cmd=f'ps -Ao "%p#%c" -ww --no-headers --sort={"-" if last_match else ""}pid',
                               match=_match)


async def find_process_by_command(patterns: Set[Pattern], last_match: bool = False) -> Optional[Tuple[int, str]]:
    def _match(line: str) -> Optional[Tuple[int, str]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                for p in patterns:
                    cmd = line_split[1].strip()
                    if p.match(cmd):
                        try:
                            return int(line_split[0]), cmd
                        except ValueError:
                            break

    return await match_syscall(cmd=f'ps -Ao "%p#%a" -ww --no-headers --sort={"-" if last_match else ""}pid',
                               match=_match)


async def find_processes_by_command(commands: Set[str], last_match: bool = False) -> Optional[Dict[str, int]]:
    matches = {}

    def _match(line: str) -> Optional[Dict[str, int]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                cmd = line_split[1].strip()
                if cmd in commands and cmd not in matches:
                    try:
                        matches[cmd] = int(line_split[0])
                    except ValueError:
                        pass

            if len(matches) == len(commands):
                return matches

    await match_syscall(cmd=f'ps -Ao "%p#%a" -ww --no-headers --sort={"-" if last_match else ""}pid', match=_match)
    return matches if matches else None


async def find_pids_by_names(names: Collection[str], last_match: bool = False) -> Optional[Dict[str, int]]:
    matches: Dict[str, int] = {}

    def _match(line: str) -> Optional[Dict[str, int]]:
        line_strip = line.strip()

        if line_strip:
            line_split = line_strip.split('#', 1)

            if len(line_split) > 1:
                current_name = line_split[1].strip()

                if current_name and current_name not in matches and current_name in names:
                    try:
                        matches[current_name] = int(line_split[0])
                    except ValueError:
                        pass

        if len(matches) == len(names):
            return matches

    await match_syscall(cmd=f'ps -Ao "%p#%c" -ww --no-headers --sort={"-" if last_match else ""}pid', match=_match)
    return matches if matches else None


async def find_commands_by_pids(pids: Set[int]) -> Optional[Dict[int, str]]:
    if pids:
        matches: Dict[int, str] = {}

        def _match(line: str) -> Optional[Dict[int, str]]:
            line_strip = line.strip()

            if line_strip:
                line_split = line_strip.split('#', 1)

                if len(line_split) > 1:
                    try:
                        current_pid = int(line_split[0])
                    except ValueError:
                        return

                    if current_pid in pids:
                        matches[current_pid] = line_split[1].strip()

            if len(matches) == len(pids):
                return matches

        await match_syscall(cmd='ps -Ao "%p#%a" -ww --no-headers', match=_match)
        return matches if matches else None


def read_current_pids() -> Set[int]:
    return {int(pid) for pid in os.listdir('/proc') if pid.isnumeric()}


async def map_pids_by_ppid() -> Optional[Dict[int, Set[int]]]:
    code, output = await async_syscall('ps -Ao "%P#%p" -ww --no-headers')

    if code == 0 and output:
        all_procs = {}
        for line in output.split('\n'):
            if line:
                line_strip = line.strip()

                if line_strip:
                    line_split = line_strip.split('#', 1)

                    if len(line_split) == 2:
                        try:
                            current_ppid, current_pid = int(line_split[0].strip()), int(line_split[1].strip())
                        except ValueError:
                            continue

                        current_children = all_procs.get(current_ppid, set())
                        current_children.add(current_pid)
                        all_procs[current_ppid] = current_children

        return all_procs


async def find_children(ppids: Set[int], ppid_map: Optional[Dict[int, Set[int]]] = None, children: Optional[List[int]] = None) -> Optional[List[int]]:
    pids_by_ppids = ppid_map if ppid_map is not None else (await map_pids_by_ppid())
    children_list = children if children is not None else []

    if pids_by_ppids:
        current_children = set()
        for pid in ppids:
            pid_children = pids_by_ppids.get(pid)

            if pid_children:
                current_children.update((p for p in pid_children if p not in children_list and p not in ppids))

        if current_children:
            await find_children(current_children, pids_by_ppids, children_list)
            children_list.extend(current_children)

    return children_list


def run_user_command(cmd: str, user_id: int, wait: bool, timeout: Optional[float] = None,
                     env: Optional[dict] = None, response: Optional[DictProxy] = None, forbidden_env_vars: Optional[Set[str]] = BAD_USER_ENV_VARS):
    args = {"args": cmd, "shell": True, "stdin": subprocess.DEVNULL,
            "stdout": subprocess.PIPE if wait else subprocess.DEVNULL,
            "stderr": subprocess.STDOUT if wait else subprocess.DEVNULL}

    if env:
        if forbidden_env_vars:
            args['env'] = {k: v for k, v in env.items() if k not in forbidden_env_vars}
        else:
            args['env'] = env

    try:
        os.setpriority(os.PRIO_PROCESS, os.getpid(), 0)  # always launch a command with nice 0
        os.setuid(user_id)

        p = subprocess.Popen(**args)

        if response is not None:
            response['pid'] = p.pid

        if timeout is not None and timeout > 0:
            p.wait(timeout)
        elif wait:
            p.wait()

        if response is not None:
            response['exitcode'] = p.returncode

            string = StringIO()
            for output in p.stdout:
                decoded = output.decode()
                string.write(decoded)

            string.seek(0)
            response['output'] = string.read()

    except Exception as e:
        if response is not None:
            response['exitcode'] = 1
            response['output'] = traceback.format_exc()


async def run_async_user_process(cmd: str, user_id: int, user_env: Optional[dict], forbidden_env_vars: Optional[Set[str]] = BAD_USER_ENV_VARS) -> Tuple[int, Optional[str]]:
    res = new_user_process_response()

    try:
        p = Process(target=run_user_command, kwargs={'cmd': cmd, 'user_id': user_id, 'wait': True, 'timeout': None, 'env': user_env,
                                                     'response': res, 'forbidden_env_vars': forbidden_env_vars})
        p.start()

        while p.is_alive():
            await asyncio.sleep(0.001)

        return res['exitcode'], res['output']
    except:
        error_msg = traceback.format_exc().replace('\n', ' ')
        return 1, error_msg


def new_user_process_response() -> DictProxy:
    proxy_res = Manager().dict()
    proxy_res['exitcode'] = None
    proxy_res['output'] = None
    proxy_res['pid'] = None
    return proxy_res




import os
import re
from logging import Logger
from typing import Optional


class InvalidConfigurationException(Exception):
    pass


def map_config_str(string: str) -> str:
    if string:
        pattern = re.compile(r'([a-zA-Z.\-_]+)\s*(=(\s*([a-zA-Z.\-_,\d:/%*]+))?)?')
        return '\n'.join(f"{prop[0]}{f'={prop[3]}' if prop[3] else ''}" for prop in pattern.findall(string) if not prop[1] or prop[2])


def read_valid_full_config(logger: Logger) -> Optional[str]:
    config_str = os.getenv('GUAPOW_CONFIG')

    if config_str:
        config_str = config_str.strip()

        if config_str:
            valid_config_str = map_config_str(config_str)

            if valid_config_str:
                return valid_config_str
            else:
                logger.error(f"Invalid 'GUAPOW_CONFIG': {config_str}")
                raise InvalidConfigurationException()


def read_additional_profile_config(logger: Logger) -> Optional[str]:
    config_str = os.getenv('GUAPOW_PROFILE_ADD')

    if config_str:
        config_str.strip()

        if config_str:
            valid_config_str = map_config_str(config_str)

            if valid_config_str:
                return valid_config_str
            else:
                logger.warning(f"Invalid 'GUAPOW_PROFILE_ADD': {config_str}")


def is_log_enabled() -> bool:
    try:
        return bool(int(os.getenv('GUAPOW_LOG', 0)))
    except ValueError:
        return False


def is_file_log() -> bool:
    try:
        return bool(int(os.getenv('GUAPOW_LOG_FILE', 0)))
    except ValueError:
        return False



import getpass
import os
from argparse import Namespace
from logging import Logger
from pathlib import Path

from guapow.cli.command import CLICommand
from guapow.common.profile import get_profile_dir
from guapow.service.optimizer.profile import OptimizationProfile, CPUSettings, ProcessSettings, IOSchedulingClass, \
    GPUSettings, CompositorSettings


class GenerateProfile(CLICommand):

    CMD = 'gen-profile'
    PROFILES = {'default', 'media', 'game', 'steam'}

    @classmethod
    def generate_default(cls, path: str) -> OptimizationProfile:
        profile = OptimizationProfile.empty(path)
        profile.cpu = CPUSettings(performance=True)
        profile.process = ProcessSettings(None)
        profile.process.nice.level = -1
        profile.process.io.ioclass = IOSchedulingClass.BEST_EFFORT
        profile.process.io.nice_level = 0
        return profile

    @classmethod
    def generate_media(cls, path: str) -> OptimizationProfile:
        profile = cls.generate_default(path)
        profile.process.nice.level = -4
        return profile

    @classmethod
    def generate_game(cls, path: str) -> OptimizationProfile:
        profile = cls.generate_media(path)
        profile.gpu = GPUSettings(performance=True)
        profile.compositor = CompositorSettings(off=True)
        profile.process.nice.watch = True  # some games change their nice level during initialization, so it is better to keep monitoring
        return profile

    @classmethod
    def generate_steam(cls, path: str) -> OptimizationProfile:
        profile = cls.generate_game(path)
        profile.steam = True
        return profile

    def __init__(self, logger: Logger):
        super(GenerateProfile, self).__init__(logger)

    def add(self, commands: object):
        cmd = commands.add_parser('gen-profile', help='It generates template profiles. Options: {}'.format(', '.join(self.PROFILES)))
        cmd.add_argument('name', choices=self.PROFILES, help="Profile name", type=str)

    def get_command(self) -> str:
        return self.CMD

    def run(self, args: Namespace) -> bool:
        profile_name = args.name
        prof_dir = get_profile_dir(os.getuid(), getpass.getuser())
        profile_path = f'{prof_dir}/{profile_name}.profile'

        if os.path.exists(profile_path):
            self._log.error(f"Profile file '{profile_path}' already exists")
            return False

        self._log.info(f"Generating profile '{profile_name}'")

        try:
            Path(prof_dir).mkdir(parents=True, exist_ok=True)
        except OSError:
            self._log.error(f"Could not make directory '{prof_dir}'")
            return False

        profile = eval(f"self.generate_{profile_name}('{profile_path}')")

        try:
            with open(profile_path, 'w+') as f:
                f.write(profile.to_file_str())

            self._log.info(f'Profile file generated: {profile_path}')
            return True
        except OSError:
            self._log.error(f"Could not generate profile file '{profile_path}'")
            return False



import asyncio
import os
import re
import shutil
from asyncio import Lock
from io import StringIO
from logging import Logger
from re import Pattern
from typing import Optional, Tuple, Dict

from guapow.common import system
from guapow.common.system import find_process_by_name, async_syscall


class MouseCursorManager:

    def __init__(self, logger: Logger, renicing: bool = True, cursor_hidden: Optional[bool] = None):
        self._log = logger
        self._lock = Lock()
        self._re_process_pattern: Optional[Pattern] = None
        self._process_name = 'unclutter'
        self._cursor_hidden = cursor_hidden
        self._renicing = renicing

    def _gen_custom_env(self, user_env: Optional[Dict[str, str]]):
        env = {**user_env} if user_env else dict(os.environ)

        display = env.get('DISPLAY', '').strip()

        if not display:
            env['DISPLAY'] = ':0'

        return env

    def _get_matching_pattern(self) -> Pattern:
        if self._re_process_pattern is None:
            self._re_process_pattern = re.compile(r'^{}$'.format(self._process_name))

        return self._re_process_pattern

    def can_work(self) -> Tuple[bool, Optional[str]]:
        if not shutil.which(self._process_name):
            return False, f"'{self._process_name}' is not installed. It will not be possible to hide the mouse cursor"

        return True, None

    async def _renice_process(self):
        pids_found = await system.find_pids_by_names(names=(self._process_name,), last_match=True)
        pid = pids_found.get(self._process_name) if pids_found else None
        if pid:
            try:
                os.setpriority(os.PRIO_PROCESS, pid, 1)
                self._log.debug(f"'{self._process_name}' reniced to '1'")
            except Exception as e:
                self._log.warning(f"Could not renice '{self._process_name}'. Exception class: {e.__class__.__name__}")
        else:
            self._log.warning(f"Could not renice '{self._process_name}': process not found")

    async def hide_cursor(self, user_request: bool, user_env: Optional[Dict[str, str]]) -> bool:
        async with self._lock:
            hidden = bool(await find_process_by_name(self._get_matching_pattern()))

            if hidden:
                self._log.warning("Mouse cursor is already hidden")

                if self._cursor_hidden is None:  # it means unclutter was initialized by a different process
                    self._cursor_hidden = False

                return False
            else:
                cmd = 'unclutter --timeout 1 -b'
                self._log.debug(f"Hiding the mouse cursor: {cmd}")
                exitcode, _ = await async_syscall(cmd, custom_env=self._gen_custom_env(user_env), return_output=False)

                if exitcode == 0:
                    self._log.info("Mouse cursor hidden")
                    self._cursor_hidden = user_request

                    if self._renicing:
                        asyncio.get_event_loop().create_task(self._renice_process())

                    return True
                else:
                    self._log.error(f"Could not hide the mouse cursor: {self._process_name} returned an unexpected code ({exitcode})")
                    return False

    async def is_cursor_hidden(self) -> Optional[bool]:
        async with self._lock:
            return self._cursor_hidden

    async def show_cursor(self) -> Optional[bool]:
        async with self._lock:
            if await find_process_by_name(self._get_matching_pattern()):
                exitcode, output = await system.async_syscall(f'killall {self._process_name}')

                if exitcode == 0:
                    self._log.info("Displaying mouse cursor")
                    self._cursor_hidden = None  # resetting initial state
                    return True
                else:
                    msg = StringIO()
                    msg.write(f"Could not display mouse cursor. Not all '{self._process_name}' instances could be killed")

                    if output:
                        msg.write(': {}'.format(' '.join(output.split('\n'))))

                    msg.seek(0)
                    self._log.error(msg.read())
                    return False
            else:
                self._log.info(f"Mouse cursor is already being displayed: '{self._process_name}' is not running")
                self._cursor_hidden = None  # resetting initial state
                return True



import os
from logging import Logger
from typing import Dict, Tuple, Optional

from guapow import __app_name__
from guapow.common.model import FileModel
from guapow.common.model_util import FileModelFiller
from guapow.common.users import is_root_user


class ProcessWatcherConfig(FileModel):

    FILE_NAME = 'watch.conf'
    FILE_MAPPING = {'interval': ('check_interval', float, None),
                    'regex.cache': ('regex_cache', bool, True),
                    'mapping.cache': ('mapping_cache', bool, True),
                    'ignored.cache': ('ignored_cache', bool, True)}

    def __init__(self, regex_cache: Optional[bool], check_interval: Optional[float], mapping_cache: Optional[bool],
                 ignored_cache: Optional[bool]):
        self.check_interval = check_interval
        self.regex_cache = regex_cache
        self.mapping_cache = mapping_cache
        self.ignored_cache = ignored_cache

    def get_output_name(self) -> str:
        pass

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def is_valid(self) -> bool:
        return all((self.is_check_interval_valid(),
                    self.regex_cache is not None,
                    self.mapping_cache is not None,
                    self.ignored_cache is not None))

    def is_check_interval_valid(self):
        return self.check_interval is not None and self.check_interval > 0

    def setup_valid_properties(self):
        if not self.is_check_interval_valid():
            self.check_interval = 1.0

        if self.regex_cache is None:
            self.regex_cache = True

        if self.mapping_cache is None:
            self.mapping_cache = False

        if self.ignored_cache is None:
            self.ignored_cache = False

    def get_file_root_node_name(self) -> Optional[str]:
        pass

    @classmethod
    def empty(cls) -> "ProcessWatcherConfig":
        return cls(None, None, None, None)

    @classmethod
    def default(cls) -> "ProcessWatcherConfig":
        instance = cls.empty()
        instance.setup_valid_properties()
        return instance

    @classmethod
    def get_file_path_by_user(cls, user_id: int, user_name: str, logger: Logger) -> Optional[str]:
        is_root = is_root_user(user_id)

        file_path = f'/etc/{__app_name__}/{cls.FILE_NAME}' if is_root else f'/home/{user_name}/.config/{__app_name__}/{cls.FILE_NAME}'

        if os.path.isfile(file_path):
            logger.info(f"Watcher configuration file '{file_path}' found")
            return file_path
        else:
            logger.warning(f"Watcher configuration file '{file_path}' not found")
            return cls.get_file_path_by_user(0, 'root', logger) if not is_root else None  # tries to retrieve it from root


class ProcessWatcherConfigReader:

    def __init__(self, filler: FileModelFiller, logger: Logger):
        self._filler = filler
        self._log = logger

    def read_valid(self, file_path: Optional[str]) -> Optional[ProcessWatcherConfig]:
        if file_path:
            self._log.debug(f"Trying to read Watcher configuration file '{file_path}'")

            try:
                with open(file_path) as f:
                    config_str = f.read().strip()

            except FileNotFoundError:
                self._log.warning(f"Watcher configuration file '{file_path}' not found. Default settings will be used instead.")
                config_str = None

            instance = ProcessWatcherConfig.empty()

            if config_str:
                self._filler.fill(instance, config_str)

            if not instance.is_valid():
                self._log.warning(f"Invalid or missing settings in Watcher configuration file '{file_path}' (default values of some properties will be used instead)")
                instance.setup_valid_properties()

            return instance
        else:
            self._log.info("Using default Watcher settings")
            return ProcessWatcherConfig.default()



import asyncio
import re
import time
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from logging import Logger
from typing import Dict, Optional, Tuple, List, Generator

import aiofiles

from guapow import __app_name__
from guapow.common import util, system, steam
from guapow.common.dto import OptimizationRequest
from guapow.common.model import CustomEnum
from guapow.common.steam import get_exe_name
from guapow.common.users import is_root_user
from guapow.service.optimizer.profile import OptimizationProfile

DELIMITER = '%'


class LauncherSearchMode(CustomEnum):
    NAME = 'n'
    COMMAND = 'c'

    @classmethod
    def map_string(cls, string: str) -> "LauncherSearchMode":
        return LauncherSearchMode.COMMAND if string.startswith('/') else LauncherSearchMode.NAME


def gen_possible_launchers_file_paths(user_id: int, user_name: str) -> Generator[str, None, None]:
    if not is_root_user(user_id):
        yield f'/home/{user_name}/.config/{__app_name__}/launchers'

    yield f'/etc/{__app_name__}/launchers'


def map_target(string: str, mapping: str, logger: Logger) -> Optional[Tuple[str, LauncherSearchMode]]:
    string_split = string.split(DELIMITER, 1)

    if len(string_split) > 1:
        mode_str = string_split[0].strip()
        mode = LauncherSearchMode.from_value(mode_str.lower())

        if not mode:
            mode = LauncherSearchMode.map_string(string)
            logger.warning(f"Invalid launcher target '{mode_str}' for mapping: {mapping}. Default type '{mode.value}' will be considered")

        return string_split[1].strip(), mode
    else:
        return string, LauncherSearchMode.map_string(string)


def map_launchers_dict(launchers: dict, logger: Logger) -> Optional[Dict[str, Tuple[str, LauncherSearchMode]]]:
    if launchers:
        res = {}
        for launcher, target in launchers.items():
            launcher_strip = launcher.strip()

            if not launcher_strip:
                continue

            target_strip = target.strip()

            if not target_strip:
                continue

            res[launcher_strip] = map_target(target_strip, f'{launcher}{DELIMITER}{target}', logger)

        return res if res else None


async def map_launchers_file(wrapper_file: str, logger: Logger) -> Optional[Dict[str, Tuple[str, LauncherSearchMode]]]:
    wrappers = {}

    async with aiofiles.open(wrapper_file) as f:
        async for line in f:
            line_strip = line.strip()

            if line_strip and not line_strip.startswith('#'):
                line_split = line_strip.split('#', 1)[0].split('=', 1)

                if len(line_split) == 2:
                    key = line_split[0].strip()

                    if key:
                        val = line_split[1].strip()

                        if val:
                            wrappers[key] = map_target(val, line_strip, logger)

    return wrappers


class LauncherMapper(ABC):

    def __init__(self, check_time: float, logger: Logger):
        self._log = logger
        self._check_time = check_time

    @abstractmethod
    async def map_pid(self, request: OptimizationRequest, profile: OptimizationProfile) -> Optional[int]:
        pass


class ExplicitLauncherMapper(LauncherMapper):
    """
    For mappings declared in the 'launchers' file
    """

    def __init__(self, wait_time: float, logger: Logger):
        super(ExplicitLauncherMapper, self).__init__(check_time=wait_time, logger=logger)

    async def _find_wrapped_process(self, wrapped_target: Tuple[str, LauncherSearchMode], launcher: str) -> Optional[int]:
        wrapped_name, search_mode = wrapped_target[0], wrapped_target[1]
        wrapped_regex = util.map_any_regex(wrapped_name)
        wrapped_regexes = {wrapped_regex}
        self._log.debug(f"Looking for mapped process with {search_mode.name.lower()} '{wrapped_name}' (launcher={launcher})")

        time_init = datetime.now()
        time_limit = time_init + timedelta(seconds=self._check_time)

        while datetime.now() < time_limit:
            if search_mode == LauncherSearchMode.COMMAND:
                wrapped_proc = await system.find_process_by_command(wrapped_regexes, last_match=True)
            else:
                wrapped_proc = await system.find_process_by_name(wrapped_regex, last_match=True)

            if wrapped_proc is not None:
                find_time = (datetime.now() - time_init).total_seconds()
                pid_found, name_found = wrapped_proc[0], wrapped_proc[1]
                self._log.info(f"Mapped process '{name_found}' ({pid_found}) found in {find_time:.2f} seconds")
                return pid_found
            else:
                await asyncio.sleep(0.001)

        find_time = (datetime.now() - time_init).total_seconds()
        self._log.warning(f"Could not find process with {search_mode.name.lower()} '{wrapped_name}' (launcher={launcher}). Timed out in {find_time:.2f} seconds")

    async def map_pid(self, request: OptimizationRequest, profile: OptimizationProfile) -> Optional[int]:
        if profile.launcher and profile.launcher.skip_mapping:
            self._log.info(f"Skipping launcher mapping for {profile.get_log_str()} (pid: {request.pid})")
            return

        if profile.launcher and profile.launcher.mapping:
            self._log.debug(f"Checking mapped launchers for {profile.get_log_str()} (pid: {request.pid})")
            launchers = map_launchers_dict(launchers=profile.launcher.mapping, logger=self._log)
        else:
            launchers = None

            for file_path in gen_possible_launchers_file_paths(request.user_id, request.user_name):
                try:
                    self._log.debug(f"Checking mapped launchers on '{file_path}' (request: {request.pid})")
                    launchers = await map_launchers_file(file_path, self._log)
                    break
                except FileNotFoundError:
                    self._log.debug(f"Launchers file '{file_path}' not found (request: {request.pid})")

        if launchers:
            file_name = request.command.split('/')[-1].strip()

            wrapped_target = launchers.get(file_name)

            if not wrapped_target:  # if there is no exact match, check any name with regex
                for name, real_name in launchers.items():
                    if '*' in name:
                        if util.map_any_regex(name).match(file_name):
                            wrapped_target = real_name
                            break

            if wrapped_target:
                return await self._find_wrapped_process(wrapped_target, file_name)

        else:
            self._log.debug("No valid launchers mapped found")


class SteamLauncherMapper(LauncherMapper):

    def __init__(self, wait_time: float, logger: Logger):
        super(SteamLauncherMapper, self).__init__(check_time=wait_time, logger=logger)

    async def map_pid(self, request: OptimizationRequest, profile: OptimizationProfile) -> Optional[int]:
        if profile.steam:
            steam_cmd = steam.get_steam_runtime_command(request.command)

            if not steam_cmd:
                self._log.warning(f'Command not from Steam: {request.command} (pid: {request.pid})')
                return

            self._log.debug(f'Steam command detected (pid: {request.pid}): {request.command}')

            proton_name_and_paths = steam.get_proton_exec_name_and_paths(steam_cmd)

            if proton_name_and_paths:
                cmd_patterns = {re.compile(r'^{}$'.format(re.escape(cmd))) for cmd in proton_name_and_paths[1:]}
            else:
                cmd_patterns = {re.compile(r'(/bin/\w+\s+)?{}'.format(re.escape(steam_cmd)))}  # native games

            cmd_logs = ', '.join(p.pattern for p in cmd_patterns)
            self._log.debug(f'Looking for a Steam process matching one of the command patterns (pid: {request.pid}): {cmd_logs}')

            time_init = datetime.now()
            time_limit = time_init + timedelta(seconds=self._check_time)
            while datetime.now() < time_limit:
                proc_found = await system.find_process_by_command(cmd_patterns, last_match=True)
                find_time = (datetime.now() - time_init).total_seconds()

                if proc_found is not None:
                    self._log.info(f"Steam process '{proc_found[1]}' ({proc_found[0]}) found in {find_time:.2f} seconds")
                    return proc_found[0]
                else:
                    await asyncio.sleep(0.001)

            find_time = (datetime.now() - time_init).total_seconds()
            self._log.warning(f'Could not find a Steam process matching command patterns (pid: {request.pid}). Search timed out in {find_time:.2f} seconds')

            if proton_name_and_paths:
                proc_name = proton_name_and_paths[0]
            else:
                proc_name = get_exe_name(steam_cmd)

            if not proc_name:
                self._log.warning(f'Name of launched Steam command could not be determined (request={request.pid}). No extra search will be performed.')
            else:
                self._log.debug(f"Trying to find Steam process by name '{proc_name}' (request: {request.pid})")
                ti = time.time()
                proc_found = await system.find_process_by_name(re.compile(r'^{}$'.format(proc_name)), last_match=True)
                tf = time.time()

                if proc_found:
                    self._log.info(f"Steam process named '{proc_found[0]}' ({proc_found[1]}) found in {tf - ti:.2f} seconds")
                    return proc_found[0]
                else:
                    self._log.warning(f'Could not find a Steam process named {proc_name} (request={request.pid})')


class LauncherMapperManager(LauncherMapper):

    def __init__(self, check_time: float, logger: Logger, mappers: Optional[List[LauncherMapper]] = None):
        super(LauncherMapperManager, self).__init__(check_time, logger)

        if mappers:
            self._sub_mappers = mappers
        else:
            self._sub_mappers = [cls(check_time, logger) for cls in LauncherMapper.__subclasses__() if cls != self.__class__]

    async def map_pid(self, request: OptimizationRequest, profile: OptimizationProfile) -> Optional[int]:
        for mapper in self._sub_mappers:
            real_id = await mapper.map_pid(request, profile)

            if real_id:
                return real_id

    def get_sub_mappers(self) -> Optional[List[LauncherMapper]]:
        if self._sub_mappers is not None:
            return [*self._sub_mappers]




import os

from setuptools import setup, find_packages

DESCRIPTION = (
    "On-demand and auto performance optimizer for Linux applications."
)

AUTHOR = "Vinícius Moreira"
AUTHOR_EMAIL = "vinicius_fmoreira@hotmail.com"
NAME = 'guapow'
URL = f'https://github.com/vinifmor/{NAME}'

file_dir = os.path.dirname(os.path.abspath(__file__))

with open(f'{file_dir}/requirements.txt') as f:
    requirements = [line.strip() for line in f.readlines() if line]


with open(f'{file_dir}/{NAME}/__init__.py') as f:
    exec(f.readlines()[4])


setup(
    name=NAME,
    version=eval('__version__'),
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    python_requires=">=3.8",
    url=URL,
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_data={NAME: ["dist/daemon/*", "dist/daemon/systemd/root/*", "dist/daemon/systemd/user/*"]},
    include_package_data=True,
    install_requires=requirements,
    test_suite="tests",
    entry_points={
        "console_scripts": [
            f"{NAME}={NAME}.runner.main:run",
            f"{NAME}-cli={NAME}.cli.main:run",
            f"{NAME}-opt={NAME}.service.optimizer.main:start",
            f"{NAME}-watch={NAME}.service.watcher.main:start"
        ]
    },
    license="zlib/libpng",
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: zlib/libpng License',
        'Operating System :: POSIX :: Linux',
        'Topic :: System',
        'Topic :: Utilities'
    ]
)



import os
from abc import ABC, abstractmethod
from enum import Enum
from io import StringIO
from typing import Tuple, Optional, Dict, List


class CustomEnum(Enum):

    @classmethod
    def names_dict(cls) -> Dict[str, "CustomEnum"]:
        return {c.name.lower(): c for c in cls}

    @classmethod
    def from_str(cls, string: str) -> Optional["CustomEnum"]:
        if string:
            final_str = string.strip().lower()

            if final_str:
                return cls.names_dict().get(final_str)

    @classmethod
    def from_value(cls, value: object) -> Optional["CustomEnum"]:
        if value is not None:
            for c in cls:
                if c.value == value:
                    return c


class FileModel(ABC):

    @abstractmethod
    def is_valid(self) -> bool:
        pass

    @abstractmethod
    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        """
        :return: a dict mapping properties from the profile file to the model's.
        The mapped tuples represent the model's property name and type.
        """
        pass

    def get_output_name(self) -> str:
        return ''

    def get_file_properties(self) -> Dict[str, str]:
        prop_map = self.get_file_mapping()
        node = self.get_file_root_node_name()
        props = {}

        if prop_map:
            for key, prop in prop_map.items():
                prop_val = getattr(self, prop[0])

                if prop_val is not None:
                    val_str = str(prop_val)

                    if isinstance(prop_val, bool):
                        val_str = val_str.lower()

                    props[f"{'{}.'.format(node) if node else ''}{key}"] = val_str

        for val in self.__dict__.values():
            if isinstance(val, FileModel):
                sub_props = val.get_file_properties()

                if node and sub_props:
                    sub_props = {f'{node}.{key}': val for key, val in sub_props.items()}

                if sub_props:
                    props.update(sub_props)

        return props

    def to_file_str(self) -> Optional[str]:
        props = self.get_file_properties()

        if props:
            string = StringIO()
            for key, val in sorted(props.items()):
                string.write(f'{key}={val}\n')

            string.seek(0)
            return string.read()

    @abstractmethod
    def get_file_root_node_name(self) -> Optional[str]:
        pass

    def get_nested_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        """
        :return: a dict with all nested file model mappings
        """
        mapping = {}

        for prop, instance in self.__dict__.items():
            if instance and isinstance(instance, FileModel):
                for file_prop, prop_data in instance.get_full_mapping().items():
                    mapping.update({file_prop: (f'{prop}.{prop_data[0]}', prop_data[1], prop_data[2])})

        return mapping

    def get_full_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        full_mapping = {}

        mapping = self.get_file_mapping()

        if mapping:
            full_mapping.update(mapping)

        nested_mapping = self.get_nested_mapping()

        if nested_mapping:
            full_mapping.update(nested_mapping)

        root_node = self.get_file_root_node_name()
        if full_mapping and root_node:
            full_mapping = {f'{root_node}.{k}': v for k, v in full_mapping.items()}

        return full_mapping

    def reset_invalid_nested_members(self):
        for prop, val in self.__dict__.items():
            if isinstance(val, FileModel):
                val.reset_invalid_nested_members()
                if not val.is_valid():
                    setattr(self, prop, None)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        for p, v in self.__dict__.items():
            if v != getattr(other, p):
                return False

        return True

    def __hash__(self):
        hash_sum = 0

        for p, v in sorted(self.__dict__.items()):
            hash_sum += hash(v)

        return hash_sum

    def __repr__(self):
        return f'{self.__class__.__name__} {self.__dict__}'


class RootFileModel(FileModel, ABC):

    def is_valid(self) -> bool:
        valid = False
        for val in self.__dict__.values():
            if isinstance(val, FileModel):
                if val.is_valid():
                    valid = True
                else:
                    return False

        return valid

    def get_file_root_node_name(self) -> Optional[str]:
        return


class ProfileFile(RootFileModel, ABC):

    def __init__(self, path: Optional[str]):
        self.path: Optional[str] = None
        self.name: Optional[str] = None
        self.set_path(path)

    def set_path(self, path: str):
        self.path = path.strip() if path is not None else None
        self.name = '.'.join(os.path.basename(self.path).split('.')[0:-1]) if self.path else None


class ScriptSettings(FileModel):

    def __init__(self, node_name: str, scripts: Optional[List[str]] = None, wait_execution: Optional[bool] = False,
                 timeout: Optional[float] = None, run_as_root: bool = False):
        self.scripts = scripts
        self.wait_execution = wait_execution
        self.timeout = timeout  # wait timeout
        self.run_as_root = run_as_root
        self._node_name = node_name
        self._file_mapping = {f"{node_name}{f'.{p}' if p else ''}": v for p, v in
                              {'': ('scripts', List[str], None), 'wait': ('wait_execution', bool, True),
                               'timeout': ('timeout', float, None), 'root': ('run_as_root', bool, True)}.items()}

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self._file_mapping

    def get_output_name(self) -> str:
        return ''

    def is_valid(self) -> bool:
        return bool(self.scripts)

    def get_file_root_node_name(self) -> Optional[str]:
        pass

    def has_valid_timeout(self) -> bool:
        return self.timeout is not None and self.timeout > 0

    def __eq__(self, other):
        if isinstance(other, ScriptSettings):
            return self.__dict__ == other.__dict__

        return False



import asyncio
import getpass
import os
import time

from guapow import __app_name__
from guapow.common.auth import read_machine_id
from guapow.common.config import read_optimizer_config
from guapow.common.log import new_logger, get_log_level
from guapow.common.model_util import FileModelFiller
from guapow.service.watcher import mapping, ignored
from guapow.service.watcher.config import ProcessWatcherConfig, ProcessWatcherConfigReader
from guapow.service.watcher.core import ProcessWatcherContext, ProcessWatcher
from guapow.service.watcher.patterns import RegexMapper


def is_log_enabled() -> bool:
    try:
        return bool(int(os.getenv('GUAPOW_WATCH_LOG', 1)))
    except ValueError:
        return True


def is_service() -> bool:
    try:
        return bool(int(os.getenv('GUAPOW_WATCH_SERVICE', 1)))
    except ValueError:
        return True


async def watch():
    ti = time.time()
    service, log_enabled = is_service(), is_log_enabled()
    logger = new_logger(name=f'{__app_name__}-watcher', service=service, enabled=log_enabled,
                        write_to_file=False, level=get_log_level('GUAPOW_WATCH_LOG_LEVEL'))
    user_id, user_name = os.getuid(), getpass.getuser()
    logger.debug(f"Initializing as system user '{user_name}'")

    filler = FileModelFiller(logger)
    watch_config_path = ProcessWatcherConfig.get_file_path_by_user(user_id, user_name, logger)
    watch_config = ProcessWatcherConfigReader(filler, logger).read_valid(watch_config_path)

    if not watch_config:
        exit(1)

    opt_config = await read_optimizer_config(user_id=user_id, user_name=user_name, filler=filler, logger=logger,
                                             only_properties={'port', 'request.encrypted'})

    if not opt_config:
        exit(1)

    machine_id = await read_machine_id() if opt_config.encrypted_requests else None

    user_env = dict(os.environ)

    if 'DISPLAY' not in user_env:
        user_env['DISPLAY'] = ':0'

    context = ProcessWatcherContext(user_id=os.getuid(), user_name=getpass.getuser(), user_env=user_env,
                                    logger=logger, optimized={}, opt_config=opt_config, watch_config=watch_config,
                                    mapping_file_path=mapping.get_default_file_path(user_id, user_name, logger),
                                    machine_id=machine_id,
                                    ignored_file_path=ignored.get_default_file_path(user_id, user_name, logger),
                                    ignored_procs=dict())

    regex_mapper = RegexMapper(cache=watch_config.regex_cache, logger=logger)
    watcher = ProcessWatcher(regex_mapper, context)
    logger.info(f'Requests encryption: {str(machine_id is not None).lower()}')
    logger.info(f'Regex cache: {str(watch_config.regex_cache).lower()}')
    logger.info(f'Mapping cache: {str(watch_config.mapping_cache).lower()}')
    logger.info(f'Ignored cache: {str(watch_config.ignored_cache).lower()}')
    logger.info(f'Checking processes every {watch_config.check_interval} second(s)')

    tf = time.time()
    logger.debug(f'Initialization took {tf - ti:.4f} seconds')
    await watcher.watch()


def start():
    asyncio.get_event_loop().run_until_complete(watch())


if __name__ == '__main__':
    start()



from abc import ABC, abstractmethod
from logging import Logger
from typing import Optional, Tuple, Dict, Set, Type

from guapow.common.dto import OptimizationRequest
from guapow.common.model import ScriptSettings
from guapow.service.optimizer.cpu import CPUFrequencyManager, CPUEnergyPolicyManager
from guapow.service.optimizer.flow import OptimizationQueue
from guapow.service.optimizer.gpu import GPUManager, GPUDriver, GPUState
from guapow.service.optimizer.mouse import MouseCursorManager
from guapow.service.optimizer.profile import OptimizationProfile
from guapow.service.optimizer.win_compositor import WindowCompositor


class OptimizationContext:

    def __init__(self, gpu_man: Optional[GPUManager], logger: Optional[Logger],
                 cpufreq_man: Optional[CPUFrequencyManager], cpuenergy_man: Optional[CPUEnergyPolicyManager],
                 mouse_man: Optional[MouseCursorManager], queue: Optional[OptimizationQueue], cpu_count: int,
                 launcher_mapping_timeout: float, renicer_interval: float,
                 compositor: Optional[WindowCompositor] = None,  allow_root_scripts: bool = False,
                 compositor_disabled_context: Optional[dict] = None,
                 system_service: bool = False, gpu_ids: Optional[Set[str]] = None):

        self.queue = queue
        self.gpu_man = gpu_man
        self.logger = logger
        self.cpufreq_man = cpufreq_man
        self.cpuenergy_man = cpuenergy_man
        self.mouse_man = mouse_man
        self.cpu_count = cpu_count
        self.compositor = compositor
        self.allow_root_scripts = allow_root_scripts
        self.launcher_mapping_timeout = launcher_mapping_timeout
        self.compositor_disabled_context = compositor_disabled_context  # if the compositor was disabled by the Optimizer
        self.renicer_interval = renicer_interval
        self.system_service = system_service
        self.gpu_ids = gpu_ids

    @classmethod
    def empty(cls) -> "OptimizationContext":
        return cls(gpu_man=None, mouse_man=None, logger=None, cpufreq_man=None, queue=None,
                   cpu_count=0, launcher_mapping_timeout=0, renicer_interval=0, cpuenergy_man=None)

    async def is_mouse_cursor_hidden(self) -> Optional[bool]:
        return await self.mouse_man.is_cursor_hidden() if self.mouse_man else None


class CPUState:

    def __init__(self, governors: Dict[str, Set[int]]):
        self.governors = governors

    def __repr__(self):
        return f'{self.__class__.__name__} {self.__dict__}'

    def __eq__(self, other):
        if isinstance(other, CPUState):
            return self.governors == other.governors

        return False

    def __hash__(self):
        return hash(self.governors)


class OptimizedProcess:

    def __init__(self, request: OptimizationRequest, created_at: float, profile: Optional[OptimizationProfile] = None,
                 previous_gpus_states: Optional[Dict[Type[GPUDriver], Set[GPUState]]] = None,
                 previous_cpu_state: Optional[CPUState] = None, stopped_after_launch: Optional[Dict[str, str]] = None,
                 cpu_energy_policy_changed: bool = False):
        self.created_at = created_at
        self.request = request
        self.profile = profile
        self.previous_gpus_states = previous_gpus_states
        self.previous_cpu_state = previous_cpu_state
        self.stopped_after_launch = stopped_after_launch
        self.alive = True
        self.related_pids = {*self.request.related_pids} if self.request and self.request.related_pids else set()
        self.pid = request.pid if self.request else None
        self.cpu_energy_policy_changed = cpu_energy_policy_changed

    def should_be_watched(self) -> bool:
        return bool(self.pid is not None and any([self.related_pids,
                                                  self.previous_cpu_state,
                                                  self.previous_gpus_states,
                                                  self.post_scripts,
                                                  self.requires_compositor_disabled,
                                                  self.stopped_processes,
                                                  self.requires_mouse_hidden,
                                                  self.stopped_after_launch,
                                                  self.cpu_energy_policy_changed]))

    @property
    def source_pid(self) -> Optional[int]:
        return self.request.pid if self.request else None

    @property
    def user_env(self) -> Optional[Dict[str, str]]:
        return self.request.user_env if self.request else None

    @property
    def user_id(self) -> Optional[int]:
        return self.request.user_id if self.request else None

    @property
    def post_scripts(self) -> Optional[ScriptSettings]:
        return self.profile.finish_scripts if self.profile else None

    @property
    def stopped_processes(self) -> Optional[Dict[str, Optional[str]]]:
        return self.request.stopped_processes if self.request else None

    @property
    def relaunch_stopped_processes(self) -> Optional[bool]:
        return self.request.relaunch_stopped_processes if self.request else None

    @property
    def requires_mouse_hidden(self) -> bool:
        return bool(self.profile and self.profile.hide_mouse)

    @property
    def requires_compositor_disabled(self) -> bool:
        return bool(self.profile and self.profile.compositor and self.profile.compositor.off)

    @property
    def relaunch_stopped_after_launch(self) -> bool:
        return bool(self.profile.stop_after.relaunch) if self.profile and self.profile.stop_after else False

    def get_display(self) -> str:
        return self.request.user_env.get('DISPLAY', ':0') if self.request and self.request.user_env else ':0'

    def get_pids(self) -> Optional[Set[int]]:
        if self.request:
            pids = set()

            if self.pid is not None:
                pids.add(self.pid)

            if self.request.pid is not None:
                pids.add(self.request.pid)

            return pids

    def __eq__(self, other):
        if not isinstance(other, OptimizedProcess):
            return False

        for p, v in self.__dict__.items():
            if v != getattr(other, p):
                return False

        return True

    def __hash__(self):
        hash_sum = 0

        for _, v in sorted(self.__dict__.items()):
            hash_sum += hash(v)

        return hash_sum

    def __repr__(self):
        return f'{self.__class__.__name__} {self.__dict__}'


class Task(ABC):

    @abstractmethod
    def __init__(self, context: OptimizationContext):
        pass

    @abstractmethod
    async def is_available(self) -> Tuple[bool, Optional[str]]:
        pass

    def is_allowed_for_self_requests(self) -> bool:
        return False

    @abstractmethod
    async def should_run(self, process: OptimizedProcess) -> bool:
        pass

    @abstractmethod
    async def run(self, process: OptimizedProcess):
        pass

